from typing import Dict, List, Optional, Tuple
import numpy as np
from itertools import chain

"""
Author: Josue N Rivera
"""
        
class Dynamics():

    def __init__(self, 
                 constants:Optional[Dict] = None,
                 state_derivative_orders:List[int] = [1],
                 control_derivative_orders:List[int] = [0]) -> None:

        self.constants = constants

        # Maximum derivitive order for each primitive state needed to represent the system's state vector (same for control)
        
        self.state_derivative_orders = np.array(state_derivative_orders, dtype=int)
        self.control_derivative_orders = np.array(control_derivative_orders, dtype=int)
        
        # Privitive state devitive order for each system state
        self.first_state_orders = np.concatenate([np.arange(i+1, dtype=int) for i in self.state_derivative_orders])
        self.first_control_orders = np.concatenate([np.arange(i+1, dtype=int) for i in self.control_derivative_orders])
        
        # One order higher of the max derivitive order for each primitive state needed to form the ode
        self.highest_state_order = max(self.state_derivative_orders)
        self.highest_control_order = max(self.control_derivative_orders)
        
        self.state_primitive_mask = np.array(list(chain(*[[True]+[False]*i for i in state_derivative_orders])), dtype=bool).reshape(-1)
        self.control_primitive_mask = np.array(list(chain(*[[True]+[False]*i for i in control_derivative_orders])), dtype=bool).reshape(-1)

        self.primitive_state_n, self.primitive_control_n = (len(state_derivative_orders), len(control_derivative_orders))

        self.first_order_state_n, self.first_order_control_n = (sum(state_derivative_orders) + self.primitive_state_n, sum(control_derivative_orders) + self.primitive_control_n)
    
    def split_first(self, first:np.ndarray):
        return tuple([first[:, i:i+1] for i in range(first.shape[1])])
    
    def f(self,
          time: np.ndarray, 
          first_order_state: np.ndarray,
          first_order_control: np.ndarray) -> np.ndarray:
        
        raise NotImplementedError
    
    def dfdx(self,
          time: np.ndarray, 
          first_order_state: np.ndarray,
          first_order_control: np.ndarray) -> np.ndarray:
        
        raise NotImplementedError
    
    def dfdu(self,
          time: np.ndarray, 
          first_order_state: np.ndarray,
          first_order_control: np.ndarray) -> np.ndarray:
        
        raise NotImplementedError
        
    def first_state_names(self) -> List[str]:
        "Returns a text label for state names"

        orders = self.state_derivative_orders

        return [f'x_{{{orders[o_idx]}}}^{{[{i}]}}' for o_idx in range(orders) for i in range(orders[o_idx]+1)]
        
    def first_control_names(self) -> List[str]:
    
        orders = self.control_derivative_orders
        return [f'u_{{{orders[o_idx]}}}^{{[{i}]}}' for o_idx in range(orders) for i in range(orders[o_idx]+1)]
        
    def first_names(self) -> Tuple[List[str], List[str]]:    
        return self.first_state_names(), self.first_control_names()
    
class WMRDynamics(Dynamics):

    """
    Dynamics for Wheeled Mobile Robot

    - Kuhne, F., Lages, W. F., & da Silva Jr, J. G. (2004, September). Model predictive control of a mobile robot using linearization. In Proceedings of mechatronics and robotics (Vol. 4, No. 4, pp. 525-530).
    - 
        
    Constants:
        wheels distance (wd; default: 1.0 m)

    """

    def __init__(self, wheels_distance=1.0) -> None:

        constants = {
            'wheels distance': wheels_distance
        }

        super().__init__(constants=constants,
                         state_derivative_orders=[0, 0, 0],
                         control_derivative_orders=[0, 0])
        
        self.wd = self.constants['wheels distance']
        
    def extract_wheel_speed(self, first_order_control):
        v, w = self.split_first(first_order_control)
        
        v_l = v - self.wd*w
        v_r = v + self.wd*w

        return np.concatenate([v_l, v_r], axis=0)

    def f(self,
          time: np.ndarray, 
          first_order_state: np.ndarray,
          first_order_control: np.ndarray) -> np.ndarray:
        
        _, _, theta = self.split_first(first_order_state)

        v, w = self.split_first(first_order_control)
        
        dx = v*np.cos(theta)
        dy = v*np.sin(theta)
        dtheta = w

        return np.concatenate([dx, dy, dtheta], axis=1)
    
    def dfdu(self, time: np.ndarray, first_order_state: np.ndarray, first_order_control: np.ndarray) -> np.ndarray:

        _, _, theta = self.split_first(first_order_state)
        v, w = self.split_first(first_order_control)

        df1 = np.concatenate([np.cos(theta), np.zeros_like(w)], axis=1)
        df2 = np.concatenate([np.sin(theta), np.zeros_like(w)], axis=1)
        df3 = np.concatenate([np.zeros_like(v), np.ones_like(w)], axis=1)

        return np.stack([df1, df2, df3], axis=1)
    
    def dfdx(self, time: np.ndarray, first_order_state: np.ndarray, first_order_control: np.ndarray) -> np.ndarray:

        x, y, theta = self.split_first(first_order_state)
        v, _ = self.split_first(first_order_control)

        df1 = np.concatenate([np.zeros_like(x), np.zeros_like(y), -v*np.sin(theta)], axis=1)
        df2 = np.concatenate([np.zeros_like(x), np.zeros_like(y), v*np.cos(theta)], axis=1)
        df3 = np.concatenate([np.zeros_like(x), np.zeros_like(y), np.zeros_like(theta)], axis=1)

        return np.stack([df1, df2, df3], axis=1)
    
class QuadcopterDynamics(Dynamics):

    """
    Quadcopter

    - Abdelhay, S., & Zakriti, A. (2019). Modeling of a quadcopter trajectory tracking system using PID controller. Procedia Manufacturing, 32, 564-571.
    """

    def __init__(self, 
                 inertia = [3.8e-3, 3.8e-3, 7.1e-3],
                 mass = 5.2,
                 gravity = 9.807,
                 arm_length = 0.32,
                 thrust_constant = 3.13e-5,
                 translational_drag = [0.1, 0.1, 0.15],
                 torque_constant = 7.5e-7,
                 rotational_drag = [0.1, 0.1, 0.15],
                 motor_inertia = 6e-5) -> None:
        
        """
        Dynamics for a quadcopter
        
        Constants:
            Inertia (I; default: [3.8*10^-3, 3.8*10^-3, 7.1*10^-3] kg m^2)
            Mass (m; default: 5.2 kg)
            Gravity (g; default: 9.807 m^2/s)
            Arm length (l; 0.32 m)

            Thrust constant (k_t; default: 3.13*10^-5 kg m)
            Net translational drag/friction coefficients (C_t; default: [0.1, 0.1, 0.15] kg s^-1)

            Torque constant (k_r; default: 7.5*10^-7 kg m)
            Net rotational drag/friction coefficients (C_r; default: [0.1, 0.1, 0.15] kg m)

            Inertia of motor (J_r; default: 6*10^-5 kg m^2) (For gyroscopic effect)

        """

        constants = {
            'inertia': inertia,
            'mass': mass,
            'gravity': gravity,
            'arm length': arm_length,
            'thrust constant': thrust_constant,
            'translational drag': translational_drag,
            'torque constant': torque_constant,
            'rotational drag': rotational_drag,
            'motor inertia': motor_inertia
        }

        super().__init__(constants,
                         state_derivative_orders=[1]*6,
                         control_derivative_orders=[0]*4,)
        
        self.m = self.constants['mass']
        self.Ct = self.constants['translational drag']
        self.g = self.constants['gravity']
        self.I = self.constants['inertia']
        self.Cr = self.constants['rotational drag']
        self.Jr = self.constants['motor inertia']

        self.b = self.constants['thrust constant']
        self.d = self.constants['torque constant']
        self.l = self.constants['arm length']

        self.w2_to_control = np.array([
            [self.b, self.b, self.b, self.b],
            [0, -self.b * self.l, 0, self.b * self.l],
            [-self.b * self.l, 0, self.b * self.l, 0],
            [self.d, -self.d, self.d, -self.d]
        ])

        self.control_to_w2 = np.linalg.inv(self.w2_to_control)

    def extract_w(self, first_order_control:np.ndarray) -> np.ndarray:

        w_squared = first_order_control @ self.control_to_w2.T

        w_squared = np.clip(w_squared, a_min=0.0, a_max=None)
        return np.sqrt(w_squared)
    
    def f(self,
          time: np.ndarray, 
          first_order_state: np.ndarray,
          first_order_control: np.ndarray) -> np.ndarray:
        
        x, dx, y, dy, z, dz, phi, dphi, theta, dtheta, psi, dpsi = tuple([first_order_state[:, i:i+1] for i in range(first_order_state.shape[1])])

        u1, u2, u3, u4 = tuple([first_order_control[:, i:i+1] for i in range(first_order_control.shape[1])])

        ux = np.cos(psi)*np.sin(theta)*np.cos(phi) + np.sin(psi)*np.sin(phi)

        uy = np.cos(phi)*np.sin(psi)*np.sin(theta) - np.cos(psi)*np.sin(phi)

        ws = self.extract_w(first_order_control)
        w1, w2, w3, w4 = tuple([ws[:, i:i+1] for i in range(ws.shape[1])])

        omega_r = w1 - w2 + w3 - w4

        return np.concatenate([
            dx,
            (u1/self.m)*ux - (self.Ct[0]/self.m)*dx,
            dy,
            (u1/self.m)*uy - (self.Ct[1]/self.m)*dy,
            dz,
            (u1/self.m)*np.cos(theta)*np.cos(phi) - (self.Ct[2]/self.m)*dz - self.g,
            dphi,
            (u2 - self.Cr[0]*dphi**2 - self.Jr*omega_r*dtheta - (self.I[2] - self.I[1])*dtheta*dpsi)/self.I[0],
            dtheta,
            (u3 - self.Cr[1]*dtheta**2 + self.Jr*omega_r*dphi - (self.I[0] - self.I[2])*dphi*dpsi)/self.I[1],
            dpsi,
            (u4 - self.Cr[2]*dpsi**2 - (self.I[1] - self.I[0])*dphi*dtheta)/self.I[2]
        ], axis=1)
    



import numpy as np
import larp

"""
Author: Josue N Rivera
"""

def test_eval():

    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[100, 0], [0, 25]]
        },
        {
            "type": "Point",
            "coordinates": [60, 55], 
            "repulsion": [[144, 0], [0, 144]]
        },
        {
            "type": "Point",
            "coordinates": [55, 48], 
            "repulsion": [[64, 0], [0, 100]]
        },
        {
            "type": "LineString",
            "coordinates": [[62, 53], [62, 60], [65, 65], [60, 60]], 
            "repulsion": [[25, 0], [0, 25]]
        }
    ]
    
    field = larp.PotentialField(size=(100, 100), rgjs=rgjs)
    
    x = np.array([[50, 65], [70, 60], [60, 60], [63, 63], [50, 50], [65, 70]])
    
    out = field.eval(x)
    assert len(np.squeeze(out)) == len(x),   "Evaluation of line string rgj does not return a size equal to the size of the input"
    assert np.squeeze(out)[0] != 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[1] != 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[2] == 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[3] == 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[4] == 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[5] == np.exp(-1), "Evaluation of line string rgj is incorrect for a point one magnitude away"

def test_area_estimation():
    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[1, 0], [0, 1]]
        }
    ]

    field = larp.PotentialField(size=(100, 100), rgjs=rgjs)
    area = field.estimate_route_area([(49, 50), (51, 50)], step=0.0001)
    assert ((area - 1.49364)**2).sum() < 1e-5, "Area estimation off"

    area = field.estimate_route_area([(49, 50), (51, 50)], step=0.0001, scale_transform=lambda x: 1/(1.0 - x + 0.000001))

def test_gradient():
    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "LineString",
            "coordinates": [[10, 10], [10, 20], [20, 20], [20, 10]], 
            "repulsion": [[2, 0], [0, 2]]
        },
        {
            "type": "Rectangle",
            "coordinates": [[30, 30], [25, 25]], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "Ellipse",
            "coordinates": [80, 80], 
            "repulsion": [[4, 0], [0, 4]],
            "shape": [[2, 0], [0, 2]]
        }
    ]

    field = larp.PotentialField(size=(100, 100), rgjs=rgjs)
    grad = field.gradient([(49, 50), (51, 50), (51, 51)])

    assert ((grad[0] - np.array([2*np.exp(-1), 0]))**2).sum() < 1e-5, "Unexpected gradient"
    assert ((grad[1] - np.array([-2*np.exp(-1), 0]))**2).sum() < 1e-5, "Unexpected gradient"
    assert ((grad[2] - np.array([-2*np.exp(-2), -2*np.exp(-2)]))**2).sum() < 1e-5, "Unexpected gradient"

    
    grad = field.gradient([(11, 10), (20, 11), (32, 31), (81, 82)])

def test_bbox():
    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "LineString",
            "coordinates": [[10, 10], [10, 20], [20, 20], [20, 10]], 
            "repulsion": [[2, 0], [0, 2]]
        },
        {
            "type": "Rectangle",
            "coordinates": [[30, 30], [25, 25]], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "Ellipse",
            "coordinates": [80, 80], 
            "repulsion": [[4, 0], [0, 4]],
            "shape": [[2, 0], [0, 2]]
        }
    ]

    field = larp.PotentialField(size=(100, 100), rgjs=rgjs)

    assert field.rgjs[1].in_bbox([15, 15]),   "Error determining bbox for linestring"
    assert field.find_bbox([81, 80])[0] == 3, "Error finding bbox for ellipse"
    assert field.find_bbox([15, 15])[0] == 1, "Error finding bbox for linestring"

test_eval()
test_area_estimation()
test_gradient()
test_bbox()


import numpy as np
import larp

"""
Author: Josue N Rivera
"""

def test_point_rgj():
    point_rgj = {
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[10, 0], [0, 5]]
    }

    point_rgj = larp.PointRGJ(coordinates = np.array(point_rgj["coordinates"]), repulsion= np.array(point_rgj["repulsion"]))

    x = np.array([[50, 65], [55, 55], [50, 50]])
    out:np.ndarray = point_rgj.eval(x)
    
    assert np.squeeze(out)[2] == 1.0, "Evaluation of point rgj is incorrect for point at origin"
    assert len(np.squeeze(out)) == 3, "Evaluation of point rgj does not return a size equal to the size of the input"

def test_line_string_rgj():
    lines_rgj = {
        'type': "LineString",
        'coordinates': [[62, 53], [62, 60], [65, 65], [60, 60]], 
        'repulsion': [[5, 0], [0, 5]]
    }

    lines_rgj = larp.LineStringRGJ(coordinates = np.array(lines_rgj["coordinates"]), repulsion = np.array(lines_rgj["repulsion"]))

    x = np.array([[62, 53], [65, 65], [60, 60], [62, 62]])

    # Test points in line string pairing
    out:np.ndarray = lines_rgj.points_in_line_pair
    assert np.all(out[0] == lines_rgj.coordinates[:2]),  "Points 1 and 2 not paired correctly as a line"
    assert np.all(out[1] == lines_rgj.coordinates[1:3]), "Points 2 and 3 not paired correctly as a line"
    assert np.all(out[2] == lines_rgj.coordinates[2:]),  "Points 3 and 4 not paired correctly as a line"

    # Test group of lines
    out:np.ndarray = lines_rgj.eval(x)
    assert len(np.squeeze(out)) == len(x), "Evaluation of line string rgj does not return a size equal to the size of the input"
    assert np.squeeze(out)[0] == 1.0, "Evaluation of line string rgj is incorrect for point at an origin"
    assert np.squeeze(out)[1] == 1.0, "Evaluation of line string rgj is incorrect for point at an origin"
    assert np.squeeze(out)[2] == 1.0, "Evaluation of line string rgj is incorrect for point at an origin"
    assert np.squeeze(out)[3] == 1.0, "Evaluation of line string rgj is incorrect for point at an origin"

def test_ellipse_rgj():
    
    rgj = larp.EllipseRGJ(coordinates=np.array([0, 0]), shape=np.eye(2), repulsion=np.eye(2))

    vectors = rgj.repulsion_vector([(0.0, 0.0), (0.0, 1.0), (1.0, 0.0), (1.0, 1.0)])

    assert ((vectors[0] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[2] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[3] - np.array([(np.sqrt(2)-1)/np.sqrt(2)]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    
    rgj = larp.EllipseRGJ(coordinates=np.array([0, 0]), shape=np.eye(2)*2, repulsion=np.eye(2))
    vectors = rgj.repulsion_vector([(1.0, 1.0), (2.0, 2.0)])

    assert ((vectors[0] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([2*(np.sqrt(2)-1)/np.sqrt(2)]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"

def test_rect_rgj():
    
    rgj = larp.RectangleRGJ(coordinates=np.array([[1]*2, [-1]*2]), repulsion=np.eye(2))

    vectors = rgj.repulsion_vector([(0.0, 0.0), (0.0, 1.0), (1.0, 0.0), (1.0, 1.0), (-2, -2)])

    assert ((vectors[0] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[2] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[3] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[4] - np.array([-1.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"

def test_multi_point_rgj():
    
    rgj = larp.MultiPointRGJ(coordinates=[(1, 1), (0, 0), (0.5, 0.5)], repulsion=np.eye(2))

    vectors = rgj.repulsion_vector([(0.15, 0.15), (0.85, 0.85), (0.45, 0.45), (0.65, 0.65)])

    assert ((vectors[0] - np.array([-0.15, -0.15]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([ 0.15,  0.15]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[2] - np.array([ 0.05,  0.05]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[3] - np.array([-0.15, -0.15]))**2).sum() < 1e-5, "Unexpected repulsion vector"

def test_multi_line_string_rgj():
    
    rgj = larp.MultiLineStringRGJ(coordinates=[[(0, 1), (0, 0)], [(1, 1), (1, 0), (2, 0)]], repulsion=np.eye(2))

    vectors = rgj.repulsion_vector([(0.15, 0.15), (0.85, 0.15), (0.45, 0.45), (0.55, 0.45)])

    assert ((vectors[0] - np.array([ 0.15, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([-0.15, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[2] - np.array([ 0.45, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[3] - np.array([-0.45, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"

def test_multi_rect_rgj():
    
    rgj = larp.MultiRectangleRGJ(coordinates=[[(0, 1), (-1, 0)], [(1, 1), (2, 0)]], repulsion=np.eye(2))

    vectors = rgj.repulsion_vector([(0.15, 0.15), (0.85, 0.15), (0.45, 0.45), (0.55, 0.45)])

    assert ((vectors[0] - np.array([ 0.15, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([-0.15, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[2] - np.array([ 0.45, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[3] - np.array([-0.45, 0.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"

def test_multi_ellipse_rgj():
    
    rgj = larp.MultiEllipseRGJ(coordinates=[(0, 0), (1, 1)], shape=np.array([np.eye(2)*0.25]*2), repulsion=np.eye(2))

    vectors = rgj.repulsion_vector([(0.15, 0.15), (0.85, 0.85), (0.45, 0.45), (0.55, 0.55)])

    assert ((vectors[0] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[2] - (1-1/np.sqrt(6.48))*np.array([ 0.45, 0.45]))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[3] - (1-1/np.sqrt(6.48))*np.array([-0.45,-0.45]))**2).sum() < 1e-5, "Unexpected repulsion vector"

def test_geometry_collection_rgj():
    
    rgj = larp.GeometryCollectionRGJ(geometries=[
         {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[36, 0], [0, 25]]
        },
        {
            "type": "Point",
            "coordinates": [60, 55], 
            "repulsion": [[36, 0], [0, 36]]
        },
        {
            "type": "Point",
            "coordinates": [45, 48], 
            "repulsion": [[64, 0], [0, 100]]
        },
        {
            "type": "LineString",
            "coordinates": [[10, 10], [80, 10], [80, 80], [10, 80], [10, 30]], 
            "repulsion": [[9, 0], [0, 9]]
        },
        {
            "type": "Rectangle",
            "coordinates": [[68, 22], [85, 0]], 
            "repulsion": [[4, 0], [0, 4]]
        },
        {
            "type": "Ellipse",
            "coordinates": [10, 80], 
            "repulsion": [[9, 0], [0, 9]],
            "shape": [[4, 0], [0, 4]]
        }
    ])

    vectors = rgj.repulsion_vector([(80.0, 10.0), (11, 81), (70.0, 80.0), (40, 8)])

    assert ((vectors[0] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[1] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[2] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected repulsion vector"
    assert ((vectors[3] - np.array([0.0,-2.0]))**2).sum() < 1e-5, "Unexpected repulsion vector"

    values = rgj.eval([(80.0, 10.0), (11, 81), (70.0, 80.0), (40, 8)])

    assert ((values[0] - 1)**2).sum() < 1e-5, "Unexpected potential field evaluation"
    assert ((values[1] - 1)**2).sum() < 1e-5, "Unexpected potential field evaluation"
    assert ((values[2] - 1)**2).sum() < 1e-5, "Unexpected potential field evaluation"
    assert ((values[3] - np.exp(-4/9))**2).sum() < 1e-5, "Unexpected potential field evaluation"

    grads = rgj.gradient([(80.0, 10.0), (11, 81), (70.0, 80.0), (40, 8)])
    grad_line = rgj.rgjs[3].gradient(np.array([(40, 8)]))[0]

    assert ((grads[0] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected gradient vector"
    assert ((grads[1] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected gradient vector"
    assert ((grads[2] - np.array([0.0]*2))**2).sum() < 1e-5, "Unexpected gradient vector"
    assert ((grads[3] - grad_line)**2).sum() < 1e-5, "Unexpected gradient vector"



from typing import List, Tuple, Union
import numpy as np
from larp.types import Point
from pyproj import CRS, Transformer

"""
Author: Josue N Rivera
"""

def route_distance(route:Union[np.ndarray, List[Point]], return_joints = False) -> Union[float, Tuple[float, np.ndarray]]:
    """
    Return distance of a path defined by a list of continuous points
    """
    route = np.array(route)

    dist = np.linalg.norm(route[1:, :] - route[:-1, :], axis=1)

    if return_joints:
        cascade = np.cumsum(dist)
        return cascade[-1], cascade
    
    return dist.sum()

def project_route(route:Union[List[Point], np.ndarray], from_crs="EPSG:3857", to_crs="EPSG:4326") -> np.ndarray:

    return project_points(points=route, from_crs=from_crs, to_crs=to_crs)

def project_points(points:Union[Point, List[Point], np.ndarray], from_crs="EPSG:3857", to_crs="EPSG:4326") -> np.ndarray:

    from_crs = CRS(from_crs)
    to_crs = CRS(to_crs)

    proj = Transformer.from_crs(crs_from=from_crs, crs_to=to_crs)
    
    points = np.array(points)
    points = np.expand_dims(points, axis=0) if points.ndim < 2 else points

    return np.stack(proj.transform(points[:,0], points[:, 1]), axis=1)

def interpolate_along_route(route:Union[List[Point], np.ndarray], step=1e-3, n=0, return_step_n = False) -> Union[np.ndarray, Tuple[np.ndarray, float, int]]:
    """
    Return a set of equally spaced points along a route
    """
    route = np.array(route)
    total_dist, joints_dist = route_distance(route, return_joints=True)
    if n <= 0:
        offset = np.arange(0.0, total_dist, step)
        n = len(offset)
    else:
        offset = np.linspace(0.0, total_dist, n)
        step = total_dist/(n-1)

    lines_idx = np.digitize(offset, joints_dist, right=True)
    relative_offset = offset - np.concatenate([[0.0], joints_dist])[lines_idx]

    line_starts = route[:-1]
    line_ends = route[1:]
    lines_diff = line_ends - line_starts
    uni_vectors = lines_diff/np.linalg.norm(lines_diff, axis=1, keepdims=True)

    points = line_starts[lines_idx] + uni_vectors[lines_idx]*relative_offset.reshape(-1, 1)

    return points if not return_step_n else (points, step, n)


from os import PathLike
from typing import Union
import numpy as np
from larp.field import PotentialField, RGJGeometry
from larp.quad import QuadTree
from pyproj import CRS, Transformer
import json
import pickle

"""
Author: Josue N Rivera
"""

def saveRGeoJSON(field:PotentialField, file:Union[str, PathLike], return_bbox=False):

    with open(file, "w", encoding='utf-8') as outfile:
        json.dump(field.toRGeoJSON(return_bbox=return_bbox), outfile)

def saveQuadTree(tree:QuadTree, file:Union[str, PathLike]):

    data = tree.toDict()

    with open(file, "wb") as outfile:
        pickle.dump(data, outfile)

def loadRGeoJSON(rgeojson: dict, size_offset = 0.0) -> PotentialField:

    features = rgeojson["features"]
    rgjs = [feature["geometry"] for feature in features]
    properties = [feature["properties"] for feature in features]

    field = PotentialField(rgjs=rgjs, properties=properties, extra_info={key: rgeojson[key] for key in rgeojson if key.lower() not in ["features", "type"]})
    field.size += size_offset*2

    return field

def loadRGeoJSONFile(file: Union[str, PathLike], size_offset = 0.0) -> PotentialField:

    with open(file=file, mode='r', encoding='utf-8') as f:
        rgeojson = json.load(f)

    return loadRGeoJSON(rgeojson, size_offset=size_offset)

def loadQuadTreeFile(file: Union[str, PathLike], size_offset = 0.0, return_field = False) -> QuadTree:

    with open(file=file, mode='rb') as f:
        data:dict = pickle.load(f)

    field = loadRGeoJSON(data.pop('field'), size_offset=size_offset)
    tree = QuadTree(field=field)
    tree.fromDict(data=data)

    if return_field:
        return tree, field

    return tree

def loadGeoJSON(geojson: Union[dict, str], size_offset = 0.0):

    """
    Converts a GeoJSON into an RGeoJSON

    _Note_: All polygons will be converted to line strings.
    """
    if isinstance(geojson, str):
        geojson = json.loads(geojson)

    def __prune_polygons__(geojson:dict):
        if geojson["type"].lower() == "polygon":

            if len(geojson["coordinates"]) == 1:
                geojson["type"] = "LineString"
                geojson["coordinates"] = geojson["coordinates"][0]
            else:
                geojson["type"] = "MultiLineString"

        elif geojson["type"].lower() == "multipolygon":
            geojson["type"] = "MultiLineString"

            coords = []
            for coord in geojson["coordinates"]:
                coords.extend(coord)
            geojson["coordinates"] = coords

        elif geojson["type"].lower() == "geometrycollection":
            for idx in range(len(geojson["geometries"])):
                __prune_polygons__(geojson["geometries"][idx])

    for idx in range(len(geojson["features"])):
        __prune_polygons__(geojson["features"][idx]["geometry"])

    return loadRGeoJSON(geojson, size_offset=size_offset)

def loadGeoJSONFile(file: Union[str, PathLike], size_offset = 0.0):
    
    with open(file=file, mode='r', encoding='utf-8') as f:
        geojson = json.load(f)

    return loadGeoJSON(geojson, size_offset=size_offset)

def projectCoordinates(field: PotentialField, from_crs="EPSG:4326", to_crs="EPSG:3857", recal_size=True):

    from_crs = CRS(from_crs)
    to_crs = CRS(to_crs)

    proj = Transformer.from_crs(crs_from=from_crs, crs_to=to_crs)

    def __prune_coords__(rgj:RGJGeometry):

        if rgj.RGJType.lower() == "geometrycollection":
            for rgj_n in rgj.rgjs:
                __prune_coords__(rgj_n)
        elif rgj.RGJType.lower() in ["point", "ellipse"]:
            rgj.set_coordinates(np.array(proj.transform(rgj.coordinates[0], rgj.coordinates[1])))
        elif rgj.RGJType.lower() in ["linestring", "rectangle", "multipoint", "multiellipse"]:
            rgj.set_coordinates(np.stack(proj.transform(rgj.coordinates[:,0], rgj.coordinates[:, 1]), axis=1))
        elif rgj.RGJType.lower() == "multirectangle":
            rgj.set_coordinates(np.stack(proj.transform(rgj.coordinates[:,:,0], rgj.coordinates[:,:,1]), axis=2))
        elif rgj.RGJType.lower() == "multilinestring":
            new_coords = []
            for coord_idx in range(len(rgj.coordinates)):
                new_coords.append(np.stack(proj.transform(rgj.coordinates[coord_idx][:, 0],\
                                                                     rgj.coordinates[coord_idx][:, 1]), axis=1))
            rgj.set_coordinates(new_coords)

    for rgj in field:
        __prune_coords__(rgj)

    if recal_size:
        field.reload_bbox()
        field.reload_center_point(toggle=True, recal_size=True)



from typing import Any, Callable, List, Tuple, TypedDict, Union, Literal
import numpy as np

"""
Author: Josue N Rivera

TODO: Add quads and eclipse
"""

FieldSize = Tuple[int, int]

Point = Union[Tuple[float, float], np.ndarray]
RepulsionVectorsAndRef = Tuple[List[int], np.ndarray]

RoutingAlgorithmStr = Literal['a*', 'dijkstra']

Scaler = Callable[[float], float]

RGJDict = TypedDict('LOIDict', { #super set of GeoJSON geometry
    'type': str,
    'coordinates': Any, 
    'repulsion': List[List[float]]
})

RGeoJSONObject = TypedDict('RGeoJSONObject', { 
    'type': Literal["Feature"],
    'properties': dict, 
    'geometry': RGJDict
})

RGeoJSONCollection = TypedDict('RGeoJSONCollection', { 
    'type': Literal["FeatureCollection"],
    'features': List[RGeoJSONObject]
})

PointRGJDict = TypedDict('PointLOIDict', {
    'type': str,
    'coordinates': List[float], 
    'repulsion': List[List[float]]
})

LineStringRGJDict = TypedDict('LineStringLOIDict',{
    'type': str,
    'coordinates': List[float], 
    'repulsion': List[List[float]]
})


import numpy as np
import larp

"""
Author: Josue N Rivera

"""

def test_quad_on_simple_pf():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=50, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  build_tree=True,
                                  minimum_length_limit=2,
                                  maximum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2))
    
    quadtree.build()

    aleaf = next(iter(quadtree.leaves))
    assert aleaf == quadtree.find_quad([aleaf.center_point])[0], "Leaf found does not match expected leaf"

    manyleaves = list(quadtree.leaves)[:30]
    quads = quadtree.find_quad([quad.center_point for quad in manyleaves])
    assert np.array([manyleaves[idx] == quads[idx] for idx in range(len(quads))]).all(), "Many quads search failed"

    assert quadtree.leaves == set(quadtree.search_leaves(quadtree.root)), "Leaves stored and leaves searched don't match"

def test_rgj_idx_passed():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)

    def get_rgj_idx(quad:larp.quad.QuadNode):

        if quad.leaf:
            return
        
        for child in quad.children:
            assert set(child.rgj_idx) <= set(quad.rgj_idx), f"Child {str(child)} rgj idxs are not a subset of {str(quad)}'s"
            assert child.boundary_zone >= quad.boundary_zone, f"Child {str(child)} boundary zone is higher than {str(quad)}'s"

        for child in quad.children:
            get_rgj_idx(child)

    get_rgj_idx(quadtree.root)

def test_leaf_none_children():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)

    for quad in quadtree.leaves:
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (in list) with non-none children"

    for quad in quadtree.search_leaves():
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (by search) with non-none children"

    assert quadtree.search_leaves() == quadtree.leaves, "Leaves in list are different than those found by search"

def test_iter_quadtree():

    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)
    
    for quad in quadtree:
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (in list) with non-none children"

    for quad in quadtree.search_leaves():
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (by search) with non-none children"

    assert quadtree.search_leaves() == quadtree.leaves, "Leaves in list are different than those found by search"

def test_quad_link():

    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)
    
    quad_chain = quadtree.find_quads_chain([[50.1, 50.1]])[0]

    assert len(quad_chain) == 4, "The expected number of quads in the chain is different than the number found"

    assert all([quad.boundary_zone == 0 for quad in quad_chain]), "The expected number of quads in the chain is different than the number found"

def test_boundary():

    quad = larp.quad.QuadNode((2, 3), 2)
    assert np.allclose(quad.get_boundaries(), np.array([1, 2, 3, 4])) , "Expected edge of quads not returned"

def test_quad_shared_edge():

    # Scenerio 1: Sharing partial edge
    quad1 = larp.quad.QuadNode((2, 3), 2)
    quad2 = larp.quad.QuadNode((4, 2), 2)
    shared_edge1 = quad1.get_shared_edge(quad2)
    shared_edge2 = quad2.get_shared_edge(quad1)

    assert np.allclose(shared_edge1, np.array([[3, 2], [3, 3]])) , "Expected edge of quads not returned"
    assert np.allclose(shared_edge1, shared_edge2) , "Expected edge of quads not returned"

    # Scenerio 2: Big box to the side of small box
    quad1 = larp.quad.QuadNode((2, 4), 2)
    quad2 = larp.quad.QuadNode((5, 4), 4)
    shared_edge1 = quad1.get_shared_edge(quad2)
    shared_edge2 = quad2.get_shared_edge(quad1)

    assert np.allclose(shared_edge1, np.array([[3, 3], [3, 5]])) , "Expected edge of quads not returned"
    assert np.allclose(shared_edge1, shared_edge2) , "Expected edge of quads not returned"

    # Scenerio 3: Big box on top of small box
    quad1 = larp.quad.QuadNode((3.5, 6.5), 3)
    quad2 = larp.quad.QuadNode((4, 4), 2)
    shared_edge1 = quad1.get_shared_edge(quad2)
    shared_edge2 = quad2.get_shared_edge(quad1)

    assert np.allclose(shared_edge1, np.array([[3, 5], [5, 5]])) , "Expected edge of quads not returned"
    assert np.allclose(shared_edge1, shared_edge2) , "Expected edge of quads not returned"

    # Scenerio 4: Not sharing an edge
    quad1 = larp.quad.QuadNode((3, 14), 2)
    quad2 = larp.quad.QuadNode((6, 11), 2)
    shared_edge1 = quad1.get_shared_edge(quad2)
    shared_edge2 = quad2.get_shared_edge(quad1)

    assert shared_edge1 is None , "Edge returned when none expected"
    assert shared_edge1 is shared_edge2, "Expected edge of quads not returned"

    # Scenerio 5: Touching at a corner
    quad1 = larp.quad.QuadNode((2, 1), 2)
    quad2 = larp.quad.QuadNode((4, 3), 2)
    shared_edge1 = quad1.get_shared_edge(quad2)
    shared_edge2 = quad2.get_shared_edge(quad1)

    assert np.allclose(shared_edge1, np.array([[3, 2], [3, 2]])) , "Expected edge of quads not returned"
    assert np.allclose(shared_edge1, shared_edge2) , "Expected edge of quads not returned"

    quad1 = larp.quad.QuadNode((4, 1), 2)
    quad2 = larp.quad.QuadNode((2, 3), 2)
    shared_edge1 = quad1.get_shared_edge(quad2)
    shared_edge2 = quad2.get_shared_edge(quad1)

    assert np.allclose(shared_edge1, np.array([[3, 2], [3, 2]])) , "Expected edge of quads not returned"
    assert np.allclose(shared_edge1, shared_edge2) , "Expected edge of quads not returned"

    # Scenerio 6: Overlap
    quad1 = larp.quad.QuadNode((102, -6), 2)
    quad2 = larp.quad.QuadNode((103, -7), 2)
    shared_edge1 = quad1.get_shared_edge(quad2)
    shared_edge2 = quad2.get_shared_edge(quad1)

    assert shared_edge1 is None , "Edge returned when none expected"
    assert shared_edge1 is shared_edge2, "Expected edge of quads not returned"

def test_quad_bbox():

    # Scenerio 1: Sharing partial edge
    quad1 = larp.quad.QuadNode((2, 3), 2)
    quad2 = larp.quad.QuadNode((4, 2), 2)

    in_bbox1 = quad1.in_bbox([[1, 2], [2, 3], [-1, -1], [4, 2], [3, 3], [3.1, 2], [2.9, 2]])
    in_bbox2 = quad2.in_bbox([[1, 2], [2, 3], [-1, -1], [4, 2], [3, 3], [3.1, 2], [2.9, 2]])

    assert np.allclose(in_bbox1, np.array([True, True, False, False, True, False, True])) , "Point misclassified as being or not inside the bounding box"
    assert np.allclose(in_bbox2, np.array([False, False, False, True, True, True, False])) , "Point misclassified as being or not inside the bounding box"
    



import larp.dynamics
import numpy as np
import scipy.integrate as integrate

"""
Author: Josue N Rivera

"""

def test_whl_dynamics():

    dynamics = larp.dynamics.WMRDynamics(1.0)

    def f(t:np.ndarray, x:np.ndarray):
        t = np.reshape(t, (-1, 1))
        x = x.reshape(-1, dynamics.first_order_state_n)
        u = np.sin(t / 2).reshape(-1, 1).repeat(2, axis=1)

        return dynamics.f(t, x, u)[0]

    t_span = (0, 10)
    t = np.linspace(*t_span, 10)
    x0 = np.array([1, 2, 1])
    
    integrate.solve_ivp(f, t_span, x0, t_eval=t)

test_whl_dynamics()

def test_quadcopter_dynamics():

    dynamics = larp.dynamics.QuadcopterDynamics()

    def f(t:np.ndarray, x:np.ndarray):
        t = np.reshape(t, (-1, 1))
        x = x.reshape(-1, dynamics.first_order_state_n)
        u = np.zeros((1, 4))
        u[:, 0] = 51

        return dynamics.f(t, x, u)[0]

    t_span = (0, 5)
    t = np.linspace(*t_span, 100)
    x0 = np.array([0]*4+[1]+[0]*7)
    
    integrate.solve_ivp(f, t_span, x0, t_eval=t)

test_quadcopter_dynamics()


import numpy as np
import larp

"""
Author: Josue N Rivera
"""

def test_eval():

    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[100, 0], [0, 25]]
        },
        {
            "type": "Point",
            "coordinates": [60, 55], 
            "repulsion": [[144, 0], [0, 144]]
        },
        {
            "type": "Point",
            "coordinates": [55, 48], 
            "repulsion": [[64, 0], [0, 100]]
        },
        {
            "type": "LineString",
            "coordinates": [[62, 53], [62, 60], [65, 65], [60, 60]], 
            "repulsion": [[25, 0], [0, 25]]
        }
    ]
    
    field = larp.PotentialField(rgjs=rgjs, size=(100, 100))
    qfield = larp.quad.QPotentailField(field)
    
    x = np.array([[50, 65], [70, 60], [60, 60], [63, 63], [50, 50], [65, 70]])
    
    out = qfield.eval(x)

    assert len(np.squeeze(out)) == len(x),   "Evaluation of line string rgj does not return a size equal to the size of the input"
    assert np.squeeze(out)[0] != 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[1] != 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[2] == 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[3] == 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[4] == 1.0,        "Evaluation of line string rgj is incorrect for point at a high-energy state"
    assert np.squeeze(out)[5] == np.exp(-1), "Evaluation of line string rgj is incorrect for a point one magnitude away"

test_eval()

def test_area_estimation():
    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[1, 0], [0, 1]]
        }
    ]

    field = larp.PotentialField(rgjs=rgjs, size=(100, 100))
    qfield = larp.quad.QPotentailField(field)

    area = qfield.estimate_route_area([(49, 50), (51, 50)], step=0.0001)
    assert ((area - 1.49364)**2).sum() < 1e-5, "Area estimation off"

    area = qfield.estimate_route_area([(49, 50), (51, 50)], step=0.0001, scale_transform=lambda x: 1/(1.0 - x + 0.000001))

test_area_estimation()

def test_gradient():
    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "LineString",
            "coordinates": [[10, 10], [10, 20], [20, 20], [20, 10]], 
            "repulsion": [[2, 0], [0, 2]]
        },
        {
            "type": "Rectangle",
            "coordinates": [[30, 30], [25, 25]], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "Ellipse",
            "coordinates": [80, 80], 
            "repulsion": [[4, 0], [0, 4]],
            "shape": [[2, 0], [0, 2]]
        }
    ]

    field = larp.PotentialField(rgjs=rgjs, size=(100, 100))
    qfield = larp.quad.QPotentailField(field)

    grad = qfield.gradient([(49, 50), (51, 50), (51, 51)])

    assert ((grad[0] - np.array([2*np.exp(-1), 0]))**2).sum() < 1e-5, "Unexpected gradient"
    assert ((grad[1] - np.array([-2*np.exp(-1), 0]))**2).sum() < 1e-5, "Unexpected gradient"
    assert ((grad[2] - np.array([-2*np.exp(-2), -2*np.exp(-2)]))**2).sum() < 1e-5, "Unexpected gradient"

    
    grad = qfield.gradient([(11, 10), (20, 11), (32, 31), (81, 82)])

test_gradient()

def test_bbox():
    rgjs = [
        {
            "type": "Point",
            "coordinates": [50, 50], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "LineString",
            "coordinates": [[10, 10], [10, 20], [20, 20], [20, 10]], 
            "repulsion": [[2, 0], [0, 2]]
        },
        {
            "type": "Rectangle",
            "coordinates": [[30, 30], [25, 25]], 
            "repulsion": [[1, 0], [0, 1]]
        },
        {
            "type": "Ellipse",
            "coordinates": [80, 80], 
            "repulsion": [[4, 0], [0, 4]],
            "shape": [[2, 0], [0, 2]]
        }
    ]

    field = larp.PotentialField(rgjs=rgjs, size=(100, 100))
    qfield = larp.QPotentailField(field)

    assert field.rgjs[1].in_bbox([15, 15]),   "Error determining bbox for linestring"
    assert qfield.find_bbox([81, 80])[0] == 3, "Error finding bbox for ellipse"
    assert qfield.find_bbox([15, 15])[0] == 1, "Error finding bbox for linestring"

test_bbox()

def test_add_remove_field():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=50, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)
    
    qfield = larp.QPotentailField(field_quadtree=quadtree)

    
    assert qfield.eval([(55, 55)])[0] != 1.0, "QPotentialField eval correct"
    assert quadtree.find_quad([(54.9, 55)])[0].boundary_zone != 0, "Boundary zone correct in quadtree"

    # Add RGJ
    point = larp.PointRGJ((55, 55), repulsion=[[10, 0], [0, 10]])
    added_idx = qfield.addRGJ(point)

    assert qfield.eval([(55, 55)])[0] == 1.0, "RGJ not added to potential field"
    assert quadtree.find_quad([(54.9, 55)])[0].boundary_zone == 0, "Boundary zone not update in quadtree"

    # Delete RGJ
    qfield.delRGJ(added_idx)
    
    assert qfield.eval([(55, 55)])[0] != 1.0, "RGJ not removed from potential field"
    assert quadtree.find_quad([(54.9, 55)])[0].boundary_zone != 0, "Boundary zone not update in quadtree"

test_add_remove_field()

def test_add_rgj_idx_passed():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)
    
    qfield = larp.QPotentailField(field_quadtree=quadtree)
    qfield.addRGJ(larp.PointRGJ((55, 55), repulsion=[[25, 0], [0, 25]]))

    def get_rgj_idx(quad:larp.quad.QuadNode):
        assert all(np.array(quad.rgj_idx) < len(field)), f"Quad's rgj indexes {quad.rgj_idx} are out of bound"

        if quad.leaf:
            return
        
        for child in quad.children:
            assert set(child.rgj_idx) <= set(quad.rgj_idx), f"Child {str(child)} rgj idxs are not a subset of {str(quad)}'s: child = {child.rgj_idx} | parent = {quad.rgj_idx}"
            assert child.boundary_zone >= quad.boundary_zone, f"Child {str(child)} boundary zone is higher than {str(quad)}'s"

        for child in quad.children:
            get_rgj_idx(child)

    get_rgj_idx(quadtree.root)

test_add_rgj_idx_passed()

def test_remove_rgj_idx_passed():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [55, 55], 
        'repulsion': [[25, 0], [0, 25]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=0.5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)
    
    qfield = larp.QPotentailField(field_quadtree=quadtree)
    qfield.delRGJ([2, 3])

    def get_rgj_idx(quad:larp.quad.QuadNode):
        assert all(np.array(quad.rgj_idx) < len(field)), f"Quad's rgj indexes {quad.rgj_idx} are out of bound"

        if quad.leaf:
            return
        
        for child in quad.children:
            assert set(child.rgj_idx) <= set(quad.rgj_idx), f"Child {str(child)} rgj idxs are not a subset of {str(quad)}'s: child = {child.rgj_idx} | parent = {quad.rgj_idx}"
            assert child.boundary_zone >= quad.boundary_zone, f"Child {str(child)} boundary zone is higher than {str(quad)}'s"
            
            get_rgj_idx(child)
        
    get_rgj_idx(quadtree.root)

test_remove_rgj_idx_passed()

def test_add_leaf_none_children():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=5,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)
    
    qfield = larp.QPotentailField(field_quadtree=quadtree)
    qfield.addRGJ(larp.PointRGJ((55, 55), repulsion=[[25, 0], [0, 25]]))

    for quad in quadtree.leaves:
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (by list) with non-none children"

    for quad in quadtree.search_leaves():
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (by search) with non-none children"

    assert quadtree.search_leaves() == quadtree.leaves, "Leaves in list are different than those found by search"

test_add_leaf_none_children()

def test_remove_leaf_none_children():
    point_rgjs = [{
        'type': "Point",
        'coordinates': [50, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [60, 50], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [50, 60], 
        'repulsion': [[5, 0], [0, 5]]
    },{
        'type': "Point",
        'coordinates': [55, 55], 
        'repulsion': [[25, 0], [0, 25]]
    }]

    field = larp.PotentialField(size=40, center_point=[55, 55], rgjs=point_rgjs)
    quadtree = larp.quad.QuadTree(field=field,
                                  minimum_length_limit=1,
                                  edge_bounds=np.arange(0.2, 0.8, 0.2),
                                  build_tree=True)
    
    qfield = larp.QPotentailField(field_quadtree=quadtree)
    qfield.delRGJ([2, 3])

    for quad in quadtree.leaves:
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (by list) with non-none children"

    for quad in quadtree.search_leaves():
        assert all([child is None for child in quad.children]) == True, f"{str(quad)} is leaf (by search) with non-none children"

    assert quadtree.search_leaves() == quadtree.leaves, "Leaves in list are different than those found by search"

test_remove_leaf_none_children()



import larp.io as lpio

"""
Author: Josue N Rivera
"""

def test_load_rgj():

    field = lpio.loadRGeoJSONFile("test/data.rgj")
    field.eval([[0.0, 0.0]])

def test_load_gj():
    field = lpio.loadGeoJSONFile("test/data.geojson")
    field.eval([[0.0, 0.0]])

def test_load_quadtree():
    quadtree = lpio.loadQuadTreeFile('test/data.quad.lp')
    field = quadtree.field
    field.eval([[55.0, 55.0]])



from __future__ import annotations
from collections.abc import Iterable
from typing import Dict, List, Optional, Set, Tuple, Union
import warnings
from collections import defaultdict

import numpy as np
from larp import PotentialField

import larp.fn as lpf
from larp.field import RGJGeometry
from larp.types import Scaler, FieldSize, Point, RGJDict

"""
Author: Josue N Rivera
Generate the quadtree from the potential field
"""

def __make_index_map__(array: Union[np.ndarray, List]) -> Dict:
    """
    Creates a dictionary mapping each unique item in the input to a unique integer index,
    preserving the order of first appearance.

    Args:
        array (Union[np.ndarray, List]): A list or array of hashable items.

    Returns:
        Dict: A mapping from unique item → index.
    """
    array = list(array)
    unique_values = list(dict.fromkeys(array))
    return dict(zip(unique_values, range(len(unique_values))))

def __deduplicate_with_index_map__(array: List) -> Tuple[List, List[int]]:

    """
    Deduplicates a list while preserving order and provides an index map
    from the original list to the unique items list.

    Useful for compressing data and remapping computations onto a reduced set
    of unique values (e.g., quads or identifiers).

    Args:
        array (List): A list of hashable items.

    Returns:
        Tuple[List, List[int]]:
            - unique_items: List of unique items in first-seen order.
            - index_map: A list of the same length as the input, where each entry
                         maps to the corresponding index in `unique_items`.
    """
    if len(array) == len(set(array)):
        # All items are already unique — identity mapping
        return array, list(range(len(array)))

    item_to_index = {}
    unique_items = []
    index_map = []

    for item in array:
        if item not in item_to_index:
            item_to_index[item] = len(unique_items)
            unique_items.append(item)
        index_map.append(item_to_index[item])

    return unique_items, index_map

class QuadTree():

    def __init__(self, field: PotentialField,
                 minimum_length_limit:float = 5.0,
                 maximum_length_limit:float = np.inf,
                 edge_bounds:Union[np.ndarray, List[float]] = np.arange(0.2, 0.8, 0.2),
                 size:Optional[float] = None,
                 conservative:bool = False,
                 build_tree:bool = True) -> None:
        
        self.field = field
        self.min_sector_size = minimum_length_limit
        self.max_sector_size = maximum_length_limit
        self.size = size if size is not None else np.max(self.field.size)

        self.edge_bounds = np.sort(np.array(edge_bounds))[::-1]
        self.n_zones = len(self.edge_bounds) + 1
        self.__zones_rad_ln = -np.log(self.edge_bounds)
        self.ZONEToMaxRANGE = np.concatenate([[1.0, 1.0], self.edge_bounds])
        self.ZONEToMinRANGE = np.concatenate([self.edge_bounds[0:1], self.edge_bounds, [0.0]])
        self.conservative = conservative

        self.root = None
        self.leaves:Set[QuadNode] = set()

        if build_tree:
            self.build()

    def __iter__(self):
        self.quad_idx = 0
        self.leaves_list = list(self.leaves)
        return self
    
    def __next__(self):
        if self.quad_idx >= len(self):
            raise StopIteration
        
        out = self.leaves_list[self.quad_idx]
        self.quad_idx += 1
        return out

    def __len__(self)->int:
        return len(self.leaves)

    def mark_leaf(self, quad:QuadNode) -> None:
        quad.leaf = True
        self.leaves.add(quad)

    def __approximated_PF_zones__(self, center_point:Point, size:float, filter_idx:Optional[List[int]] = None) -> Tuple[List[int], np.ndarray]: 
        n_rgjs = len(filter_idx)
        zones = np.ones(n_rgjs, dtype=int) * self.n_zones

        rep_vectors, refs_idxs = self.field.repulsion_vectors([center_point], filted_idx=filter_idx, min_dist_select=True, return_reference=True)

        dist_sqr = (rep_vectors*rep_vectors).sum(1)
        zone0_select = dist_sqr <= (size*size)/2.0
        zones[zone0_select] = 0

        if sum(zone0_select) < n_rgjs:
            not_zone0_select = ~zone0_select
            rgjs_idx = filter_idx[not_zone0_select]
            vectors = rep_vectors[not_zone0_select]

            vectors = vectors.reshape(-1, 2)
            uni_vectors = vectors/np.linalg.norm(vectors, axis=1, keepdims=True)

            dist_sqr = self.field.squared_dist_per(center_point - uni_vectors*(size/np.sqrt(2)), idxs=rgjs_idx).ravel()

            zones[not_zone0_select] = np.digitize(dist_sqr, self.__zones_rad_ln, right=True) + 1

        return zones, rep_vectors, refs_idxs
    
    def __build__(self, center_point:Point, size:float, filter_idx:np.ndarray) -> QuadNode:
         
        quad = QuadNode(center_point=center_point, size=size)
        filter_n = len(filter_idx)

        if filter_n:
            zones, rep_vectors, refs_idxs = self.__approximated_PF_zones__(center_point=center_point, size=size, filter_idx=filter_idx)
            quad.boundary_zone = min(zones)
            
            select = zones < self.n_zones
            quad.rgj_idx = filter_idx[select]
            quad.rgj_zones = zones[select]
        else:
            quad.boundary_zone = self.n_zones

        quad.boundary_max_range = self.ZONEToMaxRANGE[quad.boundary_zone]
        
        size2 = size/2.0
        if size <= self.max_sector_size:
            if size2 < self.min_sector_size or quad.boundary_zone == self.n_zones:
                # stop subdividing if size is too small or the active zones are too far away
                self.mark_leaf(quad)
                return quad
            if self.conservative and quad.boundary_zone > 0:
                # stop subdiving if sphere does not leave zone
                lower_range = self.ZONEToMinRANGE[quad.boundary_zone]

                select = zones == quad.boundary_zone
                vectors, refs_idxs = rep_vectors[select], refs_idxs[select]
                vectors = vectors.reshape(-1, 2)
                uni_vectors = vectors/np.linalg.norm(vectors, axis=1, keepdims=True)

                bounds_evals = self.field.eval_per(center_point + uni_vectors*(size/np.sqrt(2)), idxs=refs_idxs)
                if (bounds_evals >= lower_range).any():
                    self.mark_leaf(quad)
                    return quad

        size4 = size2/2.0
        quad['tl'] = self.__build__(center_point + np.array([-size4, size4]), size2, quad.rgj_idx)
        quad['tr'] = self.__build__(center_point + np.array([ size4, size4]), size2, quad.rgj_idx)
        quad['bl'] = self.__build__(center_point + np.array([-size4,-size4]), size2, quad.rgj_idx)
        quad['br'] = self.__build__(center_point + np.array([ size4,-size4]), size2, quad.rgj_idx)

        return quad

    def build(self) -> QuadNode:
        self.leaves:Set[QuadNode] = set()
        
        self.root = self.__build__(self.field.center_point, self.size, np.arange(len(self.field)))
        return self.root
    
    def to_boundary_lines_collection(self, margin=0.1) -> List[np.ndarray]:
        lines = [quad.to_boundary_lines(margin=margin) for quad in self.leaves]
        
        return [path for line in lines for path in line]

    def replace_branch(self, rootquad:QuadNode, child:str, branch:QuadNode):
            
            if rootquad[child] is not None:
                self.leaves -= self.search_leaves(rootquad[child])

            rootquad[child] = branch
            new_leaves = self.search_leaves(rootquad[child])
            self.leaves.update(new_leaves)

    def get_quad_maximum_range(self) -> np.ndarray:
        return np.array([quad.boundary_max_range for quad in self.leaves])

    def find_quad(self, x: Union[List['Point'], np.ndarray], max_depth: int = 1000) -> List['QuadNode']:
        """
        Efficiently finds the quad node for each point, minimizing redundant traversal.

        Args:
            x (List[Point] or np.ndarray): List of 2D points.
            max_depth (int): Maximum depth to search in the quad tree.

        Returns:
            List[QuadNode]: Quad node for each point, in input order.
        """

        x = np.atleast_2d(x).astype(float)

        n_points = len(x)
        results = [None] * n_points

        def batch_traverse(quad: 'QuadNode', point_indices: np.ndarray, depth: int):
            if quad is None or quad.leaf or depth >= max_depth:
                for idx in point_indices:
                    results[idx] = quad
                return

            # Group point indices by quadrant
            children = defaultdict(list)
            for idx in point_indices:
                dx, dy = x[idx] - quad.center_point
                if dy >= 0.0:
                    direction = 'tr' if dx >= 0.0 else 'tl'
                else:
                    direction = 'br' if dx >= 0.0 else 'bl'
                children[direction].append(idx)

            # Recurse into each child with only relevant points
            for direction, indices in children.items():
                batch_traverse(quad[direction], np.array(indices), depth + 1)

        batch_traverse(self.root, np.arange(n_points), depth=0)
        return results
    
    def find_quads_chain(self, x: Union[List['Point'], np.ndarray], max_depth: int = 1000) -> List[List['QuadNode']]:
        """
        Efficiently finds the full quad traversal chain (from root to final quad) for each point.

        Args:
            x (List[Point] or np.ndarray): List of 2D points.
            max_depth (int): Maximum depth to search in the quad tree.

        Returns:
            List[List[QuadNode]]: A list of quad chains, one per point.
        """
        x = np.atleast_2d(x).astype(float)
        n_points = len(x)
        results = [[] for _ in range(n_points)]  # chain for each point

        def batch_traverse(quad: 'QuadNode', point_indices: np.ndarray, depth: int):
            if quad is None or quad.leaf or depth >= max_depth:
                for idx in point_indices:
                    results[idx].append(quad)
                return

            # Append current quad to each point's chain
            for idx in point_indices:
                results[idx].append(quad)

            # Group points by which direction they go in
            children = defaultdict(list)
            for idx in point_indices:
                dx, dy = x[idx] - quad.center_point
                if dy >= 0.0:
                    direction = 'tr' if dx >= 0.0 else 'tl'
                else:
                    direction = 'br' if dx >= 0.0 else 'bl'
                children[direction].append(idx)

            # Recurse into each quadrant
            for direction, indices in children.items():
                batch_traverse(quad[direction], np.array(indices), depth + 1)

        batch_traverse(self.root, np.arange(n_points), depth=0)
        return results
    
    def __search_leaves__(self, quad:QuadNode):
        if quad is None: raise TypeError(f"Branch missing leaf for quad {str(quad)}")
        if quad.leaf: return [quad]

        out = []
        for child in quad.children:
            out.extend(self.__search_leaves__(child))

        return out

    def search_leaves(self, quad:Optional[QuadNode] = None) -> Set[QuadNode]:
        quad = self.root if quad is None else quad
        return set(self.__search_leaves__(quad))
    
    def get_quad_zones(self):
        return np.array([quad.boundary_zone for quad in self.leaves], dtype=int)
    
    def toDict(self):
        def __save_quad__(quad:Optional[QuadNode]) -> dict:

            if quad is None:
                return None
            
            data = {
                'center_point': quad.center_point,
                'size': quad.size,
                'leaf': quad.leaf,
                'boundary_zone': quad.boundary_zone,
                'boundary_max_range': quad.boundary_max_range,
                'rgj_idx': quad.rgj_idx,
                'rgj_zones': quad.rgj_zones,
                'children': [__save_quad__(child) for child in quad.children]
            }

            return data

        data = {
            'field': self.field.toRGeoJSON(),
            'min_sector_size': self.min_sector_size,
            'max_sector_size': self.max_sector_size,
            'size': self.size,
            'edge_bounds': self.edge_bounds,
            'n_zones': self.n_zones,
            '__zones_rad_ln': self.__zones_rad_ln,
            'ZONEToMaxRANGE': self.ZONEToMaxRANGE,
            'ZONEToMinRANGE': self.ZONEToMinRANGE,
            'conservative': self.conservative,
            'root': __save_quad__(self.root)
        }

        return data

    def fromDict(self, data:dict):

        def __load_quad__(quad_data:Optional[dict]) -> Optional[QuadNode]:

            if quad_data is None:
                return None
            
            quad = QuadNode(center_point = quad_data['center_point'],
                            size = quad_data['size'])
            
            quad.leaf = quad_data['leaf']
            quad.boundary_zone = quad_data['boundary_zone']
            quad.boundary_max_range = quad_data['boundary_max_range']
            quad.rgj_idx = quad_data['rgj_idx']
            quad.rgj_zones = quad_data['rgj_zones']
            quad.children = [__load_quad__(child) for child in quad_data['children']]

            return quad

        self.min_sector_size = data['min_sector_size'] 
        self.max_sector_size = data['max_sector_size'] 
        self.size = data['size'] 
        self.edge_bounds = data['edge_bounds'] 
        self.n_zones = data['n_zones'] 
        self.__zones_rad_ln = data['__zones_rad_ln'] 
        self.ZONEToMaxRANGE = data['ZONEToMaxRANGE'] 
        self.ZONEToMinRANGE = data['ZONEToMinRANGE']
        self.root = __load_quad__(data['root'])
        self.leaves = self.search_leaves()

    def to_image(self, return_zone=False, return_extent: bool = True) -> Union[np.ndarray, Tuple[np.ndarray, List[float]]]:
        """
        Render a top-down raster image of the quadtree zoning layout.

        Args:
            return_zone (bool): Whether to return the potential limit of each zone or the zone category.
            return_extent (bool): Whether to return the real-world coordinate extent.

        Returns:
            image (np.ndarray): 2D array with potential or zone values of sectors.
            extent (Optional[List[float]]): [xmin, xmax, ymin, ymax] real-world bounds if return_extent is True.
        """
        # Get resolution as a power of two
        resolution = 2 ** int(np.floor(np.log2(self.root.size / self.min_sector_size)))+1
        pixel_size = self.root.size / resolution

        image = np.ones((resolution, resolution), dtype=int) * self.n_zones

        half_size = self.root.size / 2.0
        lower_bound = self.root.center_point - half_size
        upper_bound = self.root.center_point + half_size

        for quad in self.leaves:
            if quad.boundary_zone == self.n_zones:
                continue

            quad_half = quad.size / 2.0
            x0 = int((quad.center_point[0] - quad_half - lower_bound[0]) / pixel_size)
            y0 = int((upper_bound[1] - (quad.center_point[1] + quad_half)) / pixel_size)
            block_size = max(1, int(np.floor(quad.size / pixel_size)))

            # Ensure bounds are within the image dimensions
            x1 = min(x0 + block_size, resolution)
            y1 = min(y0 + block_size, resolution)

            image[y0:y1, x0:x1] = quad.boundary_zone

        if not return_zone:
            image_p = self.ZONEToMaxRANGE[image]
            image_p[image == 0] = np.nan
            image = image_p

        if return_extent:
            return image, [lower_bound[0], upper_bound[0], lower_bound[1], upper_bound[1]]
        return image


    def quad_to_image(self, quad:Optional[QuadNode] = None, resolution:int = 200, margin:float = 0.0) -> np.ndarray:

        if quad is None:
            quad = self.root

        return self.field.to_image(resolution=resolution,
                                   margin=margin,
                                   center_point=quad.center_point,
                                   size=[quad.size]*2,
                                   filted_idx=quad.rgj_idx)

class QuadNode():

    chdToIdx = __make_index_map__(['tl', 'tr', 'bl', 'br'])
    nghToIdx = __make_index_map__(['tl', 't', 'tr', 'r', 'br', 'b', 'bl', 'l'])
    
    def __init__(self, center_point:Point, size:float) -> None:
        self.center_point = np.atleast_1d(center_point).astype(float)
        self.size = size
        self.leaf = False
        self.boundary_zone:int = 0
        self.boundary_max_range:float = 1.0

        self.rgj_idx = np.array([], dtype=int)
        self.rgj_zones = np.array([], dtype=int)

        self.children = [None]*len(self.chdToIdx)
        self.neighbors = [None]*len(self.nghToIdx)

    def __getitem__(self, idx:Union[str, int, tuple, list]) -> Union[QuadNode, List[QuadNode]]:

        """
        If list or tuple given, then neighbors considered. Else, children will be considered.
        """

        if isinstance(idx, (list, tuple)):
            n = len(idx)
            out = [None]*n

            for i in range(n):
                id = self.nghToIdx[idx[i]] if not isinstance(idx[i], int) else idx[i]
                out[i] = self.neighbors[id]
            return out
        
        else:
            idx = self.chdToIdx[idx] if not isinstance(idx, int) else idx
            return self.children[idx]

    def __setitem__(self, idx:Union[str, int, tuple, list], value:QuadNode) -> None:
        if isinstance(idx, (list, tuple)):
            for id in idx:
                id = self.nghToIdx[id] if not isinstance(id, int) else id
                self.neighbors[id] = value
        else:
            idx = self.chdToIdx[idx] if not isinstance(idx, int) else idx
            self.children[idx] = value

    def __lt__(self, other:QuadNode):
        return self.boundary_max_range < other.boundary_max_range
    
    def get_boundaries(self) -> np.ndarray:
        """
        Returns (xmin, ymin, xmax, ymax) boundaries of the quad.
        """
        cx, cy = self.center_point
        half = self.size / 2.0
        return np.array([cx - half, cy - half, cx + half, cy + half])
    
    def in_bbox(self, points):
        points = np.atleast_2d(points)
        xmin, ymin, xmax, ymax = self.get_boundaries()

        return (points[:, 0] >= xmin) & (points[:, 1] >= ymin) & (points[:, 0] <= xmax) & (points[:, 1] <= ymax)
    
    def get_shared_edge(self, neighbor: 'QuadNode'):
        """
        Computes the shared boundary segment between this quad and a neighboring quad.

        Returns:
            np.ndarray: A (2, 2) array representing the endpoints of the shared edge.
                - If the quads share an edge (horizontal or vertical), the endpoints of that edge are returned.
                - If the quads only touch at a corner (diagonal adjacency), the single shared point is returned twice.
                - If the quads do not touch, returns None.

        Note:
        """
        if neighbor is None:
            return None

        b1 = self.get_boundaries()   # [xmin, ymin, xmax, ymax]
        b2 = neighbor.get_boundaries()

        # Extract boundaries
        x0 = max(b1[0], b2[0])  # max(xmin1, xmin2)
        y0 = max(b1[1], b2[1])  # max(ymin1, ymin2)
        x1 = min(b1[2], b2[2])  # min(xmax1, xmax2)
        y1 = min(b1[3], b2[3])  # min(ymax1, ymax2)

        # Check for overlap in both axes
        if not np.isclose(x0, x1) and not np.isclose(y0, y1):
            return None  # No overlap

        # Otherwise, it's a shared edge
        return np.array([[x0, y0], [x1, y1]]) # [xmin, ymin, xmax, ymax]

    def to_boundary_lines(self, margin=0.1) -> Tuple[np.ndarray, np.ndarray]:
        size2 = self.size/2.0 - margin
        offset = np.array([
            [-1.0, 1.0],
            [ 1.0, 1.0],
            [ 1.0,-1.0],
            [-1.0,-1.0],
            [-1.0, 1.0],
        ]) * size2
        path = self.center_point + offset

        return path[:, 0], path[:, 1]
    
    def __str__(self) -> str:
        return f"Qd({self.center_point.tolist()}, {self.size})"

class QPotentailField(PotentialField):
    """
    A potential field optimized with a quadtree structure for efficient spatial operations.

    This class wraps a `PotentialField` with a `QuadTree`, allowing for fast spatial queries
    such as locating RGJs within a region and optimized potential field evaluations and distance search.
    All RGJs must reside within the quadtree's bounds for optimal efficiency.

    Args:
        field_quadtree (Union[PotentialField, QuadTree]): Either a `PotentialField` from which
            to build a quadtree or an existing `QuadTree` with an associated field.

    Attributes:
        field (PotentialField): The underlying potential field containing RGJs.
        quadtree (QuadTree): The quadtree that spatially indexes the field.
    """

    def __init__(self, field_quadtree: Union[PotentialField, QuadTree]):
        """
        Initialize QPotentialField with a PotentialField or a QuadTree.

        If a PotentialField is provided, a QuadTree is built from it.
        If a QuadTree is provided, its associated PotentialField is used.

        Args:
            field_quadtree (Union[PotentialField, QuadTree]): The field or associated quadtree.
        """

        if isinstance(field_quadtree, PotentialField):
            field = field_quadtree
            quadtree = QuadTree(field, minimum_length_limit=(np.max(field.size)/8)*0.9, build_tree=True)
        else:
            field = field_quadtree.field
            quadtree = field_quadtree

        self.field = field
        self.quadtree = quadtree
        self.quadtree.conservative = False

    def __getattr__(self, name):
        """
        Delegate attribute access to the wrapped PotentialField
        if the attribute is not found in QPotentialField.
        """
        try:
            return getattr(self.field, name)
        except AttributeError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}' or in wrapped field")

    def __getitem__(self, idxs:Union[int, Iterable[int]]):
        """
        Access RGJ(s) in the field by index or list of indices.

        Args:
            idxs (int or Iterable[int]): Index or iterable of indices into the RGJ list.

        Returns:
            RGJGeometry or List[RGJGeometry] or None: The selected RGJ(s), or None if invalid input.
        """

        return self.field.__getitem__(idxs)

    def __iter__(self):
        """
        Return an iterator over the RGJs in the field.
        """
        return self.field.__iter__()
    
    def __next__(self):
        """
        Return the next RGJ in the field iterator.
        """
        return self.field.__next__()

    def __len__(self)->int:
        """
        Return the number of RGJs in the field.
        """
        return len(self.field)
    
    def __group_points_by_quads_with_rgjs(self, points: np.ndarray, max_depth: int = 3):
        """
        Assign points to the deepest quad containing RGJs and group indices per quad.

        Args:
            points (np.ndarray): Array of points, shape (N, 2).
            max_depth (int): Maximum quadtree depth to traverse.

        Returns:
            unique_quads (List[QuadNode]): Unique quads with RGJs.
            quad_to_point_indices (Dict[int, List[int]]): Map from quad idx to list of point indices.
        """

        points = np.atleast_2d(points).astype(float)
        quad_chains = self.quadtree.find_quads_chain(points, max_depth=max_depth)

        final_quads = []
        for chain in quad_chains:
            for node in reversed(chain):
                if len(node.rgj_idx):
                    final_quads.append(node)
                    break
            else:
                final_quads.append(chain[0])  # fallback if no RGJs found

        unique_quads, point_to_quad_idx = __deduplicate_with_index_map__(final_quads)
        quad_to_point_indices = defaultdict(list)
        for pt_idx, quad_idx in enumerate(point_to_quad_idx):
            quad_to_point_indices[quad_idx].append(pt_idx)

        return unique_quads, quad_to_point_indices
    
    def set_all_repulsion(self, new_repulsion):
        """
        Set a new repulsion vector for all RGJs and rebuild the quadtree.
        """
        self.field.set_all_repulsion(new_repulsion=new_repulsion)
        self.quadtree.build()

    def reload_bbox(self):
        """
        Reload the bounding box of the underlying PotentialField.
        """
        self.field.reload_bbox()
    
    def reload_center_point(self, toggle=True, recal_size=False) -> Point:
        """
        Reload and optionally recalculate the center point and size of the field.

        Args:
            toggle (bool): Whether to update the center point.
            recal_size (bool): Whether to recalculate the bounding size.

        Returns:
            Point: The updated center point.
        """
        return self.field.reload_center_point(toggle=toggle, recal_size=recal_size)
    
    def get_extent(self, margin:float = 0.0) -> List[float]:
        return self.field.get_extent(margin=margin)
    
    def addField(self, new_field: PotentialField, reload_bbox=True):
        """
        Adds all RGJs from another potential field to this field and updates the quadtree accordingly.

        Args:
            new_field (PotentialField): A potential field containing RGJs to add.
            reload_bbox (bool, optional): Whether to recompute the bounding box after addition. Defaults to True.

        Returns:
            np.ndarray: Indices in the field corresponding to the newly added RGJs.
        """

        if self.quadtree.conservative:
            warnings.warn("Quadtree made not conservative")
        self.quadtree.conservative = False

        n_original = len(self.field)

        # Add RGJs to field
        
        self.field.addField(new_field=new_field, reload_bbox=reload_bbox)

        # Align new_field geometry with main field
        new_field.reload_center_point(False)
        new_field.center_point = self.field.center_point
        new_field.size = self.field.size

        # Build temporary quadtree for new_field (not conservative!)
        new_qtree = QuadTree(
            new_field,
            minimum_length_limit=self.quadtree.min_sector_size,
            maximum_length_limit=self.quadtree.max_sector_size,
            edge_bounds=self.quadtree.edge_bounds,
            size=self.quadtree.size,
            build_tree=True,
            conservative=False
        )

        # Shift all RGJ indices in new quadtree
        def update_idx(quad: QuadNode):
            if quad is None or len(quad.rgj_idx) == 0:
                return
            quad.rgj_idx += n_original
            for child in quad.children:
                update_idx(child)

        update_idx(new_qtree.root)

        # Merge new quadtree into existing one
        def update_quad(rootquad: QuadNode, newquad: Optional[QuadNode]):
            if newquad is None or newquad.boundary_zone == self.quadtree.n_zones:
                return

            # Merge zone info
            if newquad.boundary_zone < rootquad.boundary_zone:
                rootquad.boundary_zone = newquad.boundary_zone
                rootquad.boundary_max_range = newquad.boundary_max_range

            # Merge RGJ indices and zones
            if len(newquad.rgj_idx) > 0:
                rootquad.rgj_idx = np.concatenate([rootquad.rgj_idx, newquad.rgj_idx])
                rootquad.rgj_zones = np.concatenate([rootquad.rgj_zones, newquad.rgj_zones])

            # If root is leaf and new is not, convert to branch
            if rootquad.leaf and not newquad.leaf:
                self.quadtree.leaves.remove(rootquad)
                rootquad.children = [None] * len(rootquad.chdToIdx)
                rootquad.neighbors = [None] * len(rootquad.nghToIdx)
                rootquad.leaf = False

            # Recurse into children
            for child in ['tl', 'tr', 'bl', 'br']:
                nq = newquad[child]
                rq = rootquad[child]

                if rq is None:
                    self.quadtree.replace_branch(rootquad, child, nq)
                else:
                    update_quad(rq, nq)

        update_quad(self.quadtree.root, new_qtree.root)

        return np.arange(n_original, len(self.field))

    def addRGJ(self, rgj:Union[RGJDict, RGJGeometry], properties:Optional[dict] = None, reload_bbox = True, **kward) -> List[int]:
        """
        Add a single RGJ to the field and update the quadtree accordingly.

        Args:
            rgj (Union[RGJDict, RGJGeometry]): The RGJ to add.
            properties (dict, optional): Optional properties for the RGJ.
            reload_bbox (bool): Whether to reload the bounding box.

        Returns:
            List[int]: Indices of the added RGJs in the field.
        """

        if not isinstance(rgj, RGJGeometry):
            if not isinstance(rgj, dict) or "type" not in rgj:
                raise ValueError("RGJ must be an RGJGeometry.")
            
            cls = globals().get(rgj["type"] + "RGJ")
            if cls is None:
                raise ValueError(f"No RGJ class found for type {rgj['type']}")
            
            rgj = cls(properties=properties, **rgj, **kward)
        
        new_field = PotentialField([rgj])

        return self.addField(new_field=new_field, reload_bbox=reload_bbox)

    def delRGJ(self, idxs: Union[int, List[int]], reload_bbox=True, pop_field=False, pop_tree=False):
        """
        Deletes one or more RGJs from the potential field and updates the quadtree to reflect these deletions.

        This involves:
        - Removing the RGJs from the field.
        - Traversing and cleaning the quadtree to update indices, remove deleted references, 
        and merge empty branches if possible.
        - Optionally returning the deleted RGJs as a field or quadtree for backup or reuse.

        Args:
            idxs (int or List[int]): Index or indices of RGJs to remove.
            reload_bbox (bool): Whether to recompute the bounding box of the field. Defaults to True.
            pop_field (bool): If True, return a PotentialField containing the removed RGJs.
            pop_tree (bool): If True, return a QuadTree built on the removed RGJs.

        Returns:
            Optional[PotentialField, QuadTree, Tuple]: Depending on `pop_field` and `pop_tree`, 
                returns the removed RGJs as a PotentialField, a QuadTree, or both.
        """
        if self.quadtree.conservative:
            warnings.warn("Quadtree made non-conservative")
            self.quadtree.conservative = False

        idxs = np.atleast_1d(idxs).astype(int)
        idxs = np.unique(idxs % len(self))[::-1]  # Wrap, deduplicate, and reverse sort

        # Step 1: Store deleted RGJs before removing from field
        rgjs = [self.field.rgjs[idx] for idx in idxs]
        self.field.delRGJ(idxs, reload_bbox=reload_bbox)

        # Step 2: Build shift map (old index -> new index)
        total = len(self.field.rgjs) + len(idxs)
        shift_map = np.arange(total)
        
        deleted = np.zeros(total, dtype=bool)
        deleted[idxs] = True
        shift_map = shift_map - np.cumsum(deleted)

        def shift_recursive(quad: QuadNode):
            if quad is None or quad.rgj_idx.size == 0:
                return
            
            quad.rgj_idx = shift_map[quad.rgj_idx]
            if not quad.leaf:
                for child in ['tl', 'tr', 'bl', 'br']:
                    shift_recursive(quad[child])

        # Step 3: Traverse quadtree and update nodes
        def clean_quad(quad: QuadNode):
            if quad is None:
                return

            # --- Filter out deleted RGJ indices ---
            keep_mask = ~np.isin(quad.rgj_idx, idxs, assume_unique=True)

            # If no RGJs in this quad are being deleted, skip processing
            if np.all(keep_mask):
                shift_recursive(quad)
                return

            # --- Apply filtering ---
            quad.rgj_idx = shift_map[quad.rgj_idx[keep_mask]]
            quad.rgj_zones = quad.rgj_zones[keep_mask]

            quad.boundary_zone = (min(quad.rgj_zones) if len(quad.rgj_zones) > 0 else self.quadtree.n_zones)

            # --- Attempt early merge (before recursion) ---
            if (not quad.leaf) and quad.size <= self.quadtree.max_sector_size and quad.boundary_zone == self.quadtree.n_zones:
                self.quadtree.leaves -= self.quadtree.search_leaves(quad)
                quad.children = [None] * len(quad.chdToIdx)
                quad.neighbors = [None] * len(quad.nghToIdx)
                self.quadtree.mark_leaf(quad)
                return

            # --- Recurse if not a leaf ---
            if not quad.leaf and keep_mask.size > 0:
                for child in ['tl', 'tr', 'bl', 'br']:
                    clean_quad(quad[child])

        clean_quad(self.quadtree.root)

        # Step 4: Optionally return removed data
        if pop_field or pop_tree:
            search_field = PotentialField(rgjs)
            search_field.reload_center_point(False)
            search_field.center_point = self.field.center_point
            search_field.size = self.field.size

            if pop_field and not pop_tree:
                return search_field

            search_qtree = QuadTree(
                search_field,
                minimum_length_limit=self.quadtree.min_sector_size,
                edge_bounds=self.quadtree.edge_bounds,
                size=self.quadtree.size,
                build_tree=True
            )

            if pop_tree and not pop_field:
                return search_qtree

            return search_field, search_qtree

    def in_bbox(self, point: Point, max_depth: int = 3, filted_idx: Optional[List[int]] = None) -> bool:
        """
        Check if the point lies within the bounding box of any RGJ using the quadtree.

        Args:
            point (Point): The 2D point to check.
            max_depth (int): Maximum depth to search in the quadtree.

        Returns:
            bool: True if the point lies in any RGJ bounding box.
        """
        if filted_idx is not None:
            return self.field.in_bbox(point=point, filted_idx=filted_idx)

        point = np.array(point, dtype=np.float64)
        quad_chain = self.quadtree.find_quads_chain([point], max_depth=max_depth)[0]

        searched_rgj = set()

        for quad in reversed(quad_chain):
            to_search_rgj = set(quad.rgj_idx) - searched_rgj

            if len(to_search_rgj):
                searched_rgj.update(to_search_rgj)

                bbox_idx = self.field.find_bbox(point, filted_idx=to_search_rgj)

                if len(bbox_idx):
                    return bbox_idx

        # fallback: search entire field if no relevant RGJs found
        return self.field.in_bbox(point)
    
    def find_bbox(self, point: Point, max_depth: int = 3, filted_idx: Optional[List[int]] = None) -> np.ndarray:
        """
        Return indices of RGJs whose bounding boxes contain the point, using the quadtree.

        Args:
            point (Point): The 2D point to check.
            max_depth (int): Maximum depth of the quadtree to consider.

        Returns:
            np.ndarray: Indices of RGJs (global) whose bounding boxes contain the point.

        Note function may return only a subset of the RGJs indexes (those closest in distance)
        """
        if filted_idx is not None:
            return self.field.find_bbox(point=point, filted_idx=filted_idx)

        point = np.array(point, dtype=np.float64)
        quad_chain = self.quadtree.find_quads_chain([point], max_depth=max_depth)[0]

        searched_rgj = set()

        for quad in reversed(quad_chain):
            to_search_rgj = set(quad.rgj_idx) - searched_rgj

            if len(quad.rgj_idx):
                searched_rgj.update(to_search_rgj)

                bbox_idx = self.field.find_bbox(point, filted_idx=to_search_rgj)

                if len(bbox_idx):
                    return bbox_idx

        # fallback: search whole field if no relevant RGJs found; not expected to ever happen
        return self.field.find_bbox(point)
    
    def repulsion_vectors(
        self,
        points: Union[np.ndarray, List['Point']],
        min_dist_select: bool = True,
        return_reference: bool = False,
        max_depth: int = 3,
        filted_idx:Optional[List[int]] = None
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """
        Computes repulsion vectors for 2D points using only relevant RGJs from a quadtree field.
        Each point is assigned to the deepest quad with non-empty RGJs, and repulsion vectors
        are calculated using those RGJ indices. Redundant calculations are avoided by grouping
        points per quad.

        Args:
            points (Union[np.ndarray, List[Point]]): Input 2D points.
            min_dist_select (bool): Whether to use minimum distance RGJ filtering.
            return_reference (bool): If True, returns RGJ indices used per point.
            max_depth (int): Maximum depth in the quadtree to traverse.

        Returns:
            If return_reference is False:
                np.ndarray of shape (N, 2): Repulsion vectors for all input points.
                
            If return_reference is True:
                Tuple[np.ndarray, np.ndarray]: (repulsion_vectors, rgj_indices_used_per_point)
        """
        if filted_idx is not None:
            return self.field.repulsion_vectors(points=points, filted_idx=filted_idx, min_dist_select=min_dist_select, return_reference=return_reference)

        points = np.atleast_2d(points).astype(float)
        n_points = len(points)
        repulsion_vectors = np.zeros((n_points, 2), dtype=float)
        rgj_reference_ids = np.zeros(n_points, dtype=int) if return_reference else None

        unique_quads, quad_to_point_indices = self.__group_points_by_quads_with_rgjs(points, max_depth=max_depth)

        for quad_idx, pt_indices in quad_to_point_indices.items():
            quad = unique_quads[quad_idx]
            rgj_indices = quad.rgj_idx

            group_points = points[pt_indices]
            if return_reference:
                group_vectors, group_rgj_ids = self.field.repulsion_vectors(
                    points=group_points,
                    filted_idx=rgj_indices,
                    min_dist_select=min_dist_select,
                    return_reference=True
                )
                repulsion_vectors[pt_indices] = group_vectors
                rgj_reference_ids[pt_indices] = group_rgj_ids
            else:
                group_vectors = self.field.repulsion_vectors(
                    points=group_points,
                    filted_idx=rgj_indices,
                    min_dist_select=min_dist_select,
                    return_reference=False
                )
                repulsion_vectors[pt_indices] = group_vectors

        if return_reference:
            return repulsion_vectors, rgj_reference_ids
        else:
            return repulsion_vectors
        
    def gradient(self, points, min_dist_select=True, max_depth=2):
        points = np.atleast_2d(points).astype(float)
        n_points = len(points)

        if len(self.field) == 0:
            return np.zeros((n_points, 2), dtype=float)

        grad = np.zeros((n_points, 2), dtype=float)

        unique_quads, quad_to_point_indices = self.__group_points_by_quads_with_rgjs(points, max_depth=max_depth)

        for quad_idx, pt_indices in quad_to_point_indices.items():
            quad = unique_quads[quad_idx]
            points_in_quad = points[pt_indices]

            for rgj_idx in quad.rgj_idx:
                grad_vecs = self.field.rgjs[rgj_idx].gradient(points_in_quad, min_dist_select=min_dist_select)
                grad[pt_indices] += grad_vecs

        return grad

    def eval(self, points: Union[np.ndarray, List['Point']], max_depth=2, filted_idx:Optional[List[int]] = None) -> np.ndarray:
        """
        Evaluate the potential field at given points using quadtree
        to filter relevant RGJs efficiently.

        Args:
            points (Union[np.ndarray, List[Point]]): Points to evaluate (Nx2).

        Returns:
            np.ndarray: Evaluated potential values at each point.
        """
        if filted_idx is not None:
            return self.field.eval(points=points, filted_idx=filted_idx)

        points = np.atleast_2d(points).astype(float)
        n_points = len(points)

        if not len(self.field.rgjs):
            return np.zeros(n_points, dtype=float)

        unique_quads, quad_to_point_indices = self.__group_points_by_quads_with_rgjs(points, max_depth=max_depth)

        results = np.zeros(n_points, dtype=float)

        for quad_idx, pt_indices in quad_to_point_indices.items():
            quad = unique_quads[quad_idx]
            rgj_indices = quad.rgj_idx
            group_points = points[pt_indices]

            if not len(rgj_indices):
                results[pt_indices] = 0.0
            else:
                results[pt_indices] = self.field.eval(group_points, filted_idx=rgj_indices)

        return results
    
    def eval_per(self, points: Union[np.ndarray, List[Point]], idxs:Optional[List[int]] = None) -> np.ndarray:
        return self.field.eval_per(points=points, idxs=idxs)
    
    def squared_dist(self, points:Union[np.ndarray, List[Point]], scaled=True, inverted=True, max_depth=2, return_reference = False, filted_idx:Optional[List[int]] = None) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:

        points = np.array(points)
        if not len(self):
            warnings.warn("There are not any RGJs elements in the field")
            if return_reference:
                return points.sum(1)*np.inf, -np.ones_like(points.sum(1))
            return points.sum(1)*np.inf

        dists = self.squared_dist_list(points=points, scaled=scaled, inverted=inverted, max_depth=max_depth, filted_idx=filted_idx)

        if return_reference:
            min_idxs = np.argmin(dists, axis=1)
            return dists[np.arange(len(dists)), min_idxs], min_idxs

        return np.min(dists, axis=1)
    
    def squared_dist_per(self, points: Union[np.ndarray, List[Point]], idxs:Optional[List[int]] = None, scaled=True, inverted=True) -> np.ndarray:
        return self.field.squared_dist_per(points=points, idxs=idxs, scaled=scaled, inverted=inverted)
    
    def squared_dist_list(
        self,
        points: Union[np.ndarray, List['Point']],
        scaled: bool = True,
        inverted: bool = True,
        max_depth: int = 3,
        filted_idx:Optional[List[int]] = None
    ) -> np.ndarray:
        """
        Compute squared distances from each point to all RGJs in the field, but only
        evaluate the ones relevant per point using the quadtree. Non-relevant RGJs will
        be set to np.inf to maintain consistent output shape.

        Args:
            points (Union[np.ndarray, List[Point]]): Input points, shape (N, 2).
            scaled (bool): Whether to scale distances.
            inverted (bool): Whether to invert distance values.
            max_depth (int): Max depth of quadtree search.

        Returns:
            np.ndarray: Squared distances of shape (N, M), where M = total RGJs.
        """
        if filted_idx is not None:
            return self.field.squared_dist_list(points=points, filted_idx=filted_idx, scaled=scaled, inverted=inverted)

        points = np.atleast_2d(points).astype(float)
        n_points = len(points)
        total_rgjs = len(self.field.rgjs)

        if total_rgjs == 0:
            warnings.warn("There are no RGJs in the field.")
            return np.ones((n_points, 1)) * np.inf

        dist_matrix = np.ones((n_points, total_rgjs), dtype=np.float64) * np.inf

        # Group by quads that contain RGJs
        unique_quads, quad_to_point_indices = self.__group_points_by_quads_with_rgjs(points, max_depth=max_depth)

        for quad_idx, pt_indices in quad_to_point_indices.items():
            quad = unique_quads[quad_idx]
            rgj_indices = quad.rgj_idx
            if not rgj_indices:
                continue

            group_points = points[pt_indices]
            group_dists = np.stack([
                self.field.rgjs[i].squared_dist(group_points, scaled=scaled, inverted=inverted)
                for i in rgj_indices
            ], axis=1)

            for i, pt_i in enumerate(pt_indices):
                for j, rgj_i in enumerate(rgj_indices):
                    dist_matrix[pt_i, rgj_i] = group_dists[i, j]

        return dist_matrix

    def estimate_route_area(self, route:Union[List[Point], np.ndarray], step=1e-3, n=0, scale_transform:Scaler = lambda x: x, max_depth:int = 3) -> float:
        route = np.array(route)

        points, step, _ = lpf.interpolate_along_route(route=route, step=step, n=n, return_step_n=True)
        points = points if n <= 0 else points[:-1]

        f_eval = scale_transform(self.eval(points=points, max_depth=max_depth))

        return f_eval.sum()*step
    
    def estimate_route_highest_potential(self, route:Union[List[Point], np.ndarray], step=1e-2, n=0, scale_transform:Scaler = lambda x: x, max_depth:int = 3) -> float:
        route = np.array(route)

        points, step, _ = lpf.interpolate_along_route(route=route, step=step, n=n, return_step_n=True)
        points = points if n <= 0 else points[:-1]

        f_eval:np.ndarray = scale_transform(self.eval(points=points, max_depth=max_depth))

        return f_eval.max()

    def to_image(self, resolution:int = 400, margin:float = 0.0, center_point:Optional[Point] = None, size:Optional[FieldSize] = None, max_depth:int = 3, return_extent=True) -> np.ndarray:

        if center_point is None:
            if self.field.center_point is None:
                raise RuntimeError('center point for field has not been defined')
            
            center_point = self.field.center_point

        if size is None:
            if self.field.size is None:
                raise RuntimeError('size of field has not been defined')

            size = self.field.size
        else:
            size = np.array(size)

        n2 = size/2.0

        loc_tl = np.array(center_point) + np.array([-n2[0]-margin, n2[1]+margin])
        loc_br = np.array(center_point) + np.array([n2[0]+margin, -n2[1]-margin])

        y_resolution = int(resolution*abs(loc_tl[1] - loc_br[1])/abs(loc_br[0] - loc_tl[0]))
        xaxis = np.linspace(loc_tl[0], loc_br[0], resolution)
        yaxis = np.linspace(loc_tl[1], loc_br[1], y_resolution)

        xgrid, ygrid = np.meshgrid(xaxis, yaxis)
        points = np.vstack([xgrid.ravel(), ygrid.ravel()]).T

        image = self.eval(points, max_depth=max_depth).reshape((y_resolution, resolution))

        if return_extent:
            extent = np.reshape([loc_tl[0], loc_br[0], loc_br[1], loc_tl[1]], -1).tolist()
            return image, extent

        return image
    



import numpy as np
import larp

"""
Author: Josue N Rivera

"""

def test_route_distance():
    dist = larp.route_distance([(0, 0), (1, 1)])

    assert ((dist - np.sqrt(2))**2).sum() < 1e-10, "Distance incorrect for diagonal line"

    dist, joints_dist = larp.route_distance([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)], return_joints=True)

    assert ((dist - 4)**2).sum() < 1e-10, "Distance incorrect for a square"
    assert ((joints_dist - np.array([1, 2, 3, 4]))**2).sum() < 1e-10, "Joint distance incorrect for a square"

def test_interpolate_route():
    points = larp.interpolate_along_route([(0, 0), (1, 1)], step=0.5)

    assert (points[1] != np.array([0.5, 0.5])).all(), "Middle point interpolated using wrong projection"
    assert len(points) == 3, "Number of points not expected"

    points = larp.interpolate_along_route([(0, 0), (1, 1)], step=1.0/np.sqrt(2))

    assert ((points[1] - np.array([0.5, 0.5]))**2).sum() < 1e-10, "Middle point interpolated incorrectly"
    assert len(points) == 3, "Number of points not expected (3 due to rounding error)"

    points, step, n = larp.interpolate_along_route([(0, 0), (1, 1)], n = 3, return_step_n=True)
    assert (points[1] == np.array([0.5, 0.5])).all(), "Middle point interpolated incorrectly with n parameter"
    assert n == 3, "Returned n not expected"
    assert ((step - 1.0/np.sqrt(2))**2).sum() < 1e-10, "Returned step not expected"
    assert len(points) == 3, "Number of points not expected"
    
if __name__ == "__main__":
    test_route_distance()
    test_interpolate_route()


from __future__ import annotations
from collections.abc import Iterable
from typing import List, Optional, Tuple, Union
import warnings

import numpy as np
import larp.fn as lpf
from larp.types import Scaler, RGJDict, FieldSize, Point, RGeoJSONCollection, RGeoJSONObject, RepulsionVectorsAndRef

"""
Author: Josue N Rivera

x are assumed to be a list of point coordinates in euclidean space

"""

class RGJGeometry():

    RGJType = None

    def __init__(self, coordinates:Union[np.ndarray, List[Point], List[List[Point]], Point], repulsion:Optional[np.ndarray] = None, properties:Optional[dict] = None, optional_dim = 2, **kwargs) -> None:
        self.coordinates = np.array(coordinates)
        self.repulsion = np.eye(optional_dim) if repulsion is None else np.array(repulsion)
        self.inv_repulsion = np.linalg.inv(self.repulsion)
        self.eye_repulsion = np.eye(len(self.repulsion))
        self.properties = {} if properties is None else properties
        self.grad_matrix = self.inv_repulsion + self.inv_repulsion.T
        
        bbox = self.coordinates.reshape(-1, 2)
        self.bbox = np.array([bbox.min(0), bbox.max(0)])

    def set_coordinates(self, new_coords):
        self.coordinates = np.array(new_coords)
        bbox = self.coordinates.reshape(-1, 2)
        self.bbox = np.array([bbox.min(0), bbox.max(0)])

    def set_repulsion(self, new_repulsion):
        self.repulsion = np.array(new_repulsion)
        self.inv_repulsion = np.linalg.inv(self.repulsion)
        self.eye_repulsion = np.eye(len(self.repulsion))
        self.grad_matrix = self.inv_repulsion + self.inv_repulsion.T

    def get_dist_matrix(self, scaled=True, inverted=True):

        if inverted and scaled:
            return self.inv_repulsion
        if not scaled:
            return self.eye_repulsion
        
        return self.repulsion

    def get_center_point(self) -> np.ndarray:
        if len(self.coordinates.shape) <= 1:
            return self.coordinates
        
        coords = np.reshape(self.coordinates, (-1, 2))

        return (coords.min(0) + coords.max(0))/2.0
    
    def in_bbox(self, x:Point) -> bool:

        bboxes = self.bbox.reshape((-1, 2))
        for i in range(0, len(bboxes), 2):
            if all(x >= bboxes[i]) and all(x <= bboxes[i+1]):
                return True

        return False

    def squared_dist(self, x: np.ndarray, scaled=True, inverted=True, **kwargs) -> np.ndarray:
        nvector = self.repulsion_vector(x, min_dist_select = True)
        matrix = self.get_dist_matrix(scaled=scaled, inverted=inverted)

        return ((nvector@matrix)*nvector).sum(axis=1)
    
    def repulsion_vector(self, x:np.ndarray, **kwargs) -> np.ndarray:
        raise NotImplementedError
    
    def gradient(self, x:np.ndarray, **kwargs):
        repulsion_vector = self.repulsion_vector(x, **kwargs)
        return - self.eval(x=x).reshape(-1, 1) * (repulsion_vector@self.grad_matrix.T)

    def eval(self, x:np.ndarray):
        return np.exp(-self.squared_dist(x))
    
    def toRGeoJSON(self) -> RGeoJSONObject:
        if self.RGJType is None: 
            return UserWarning(f"Object doesn't have a RGJType")
        
        return {
            "type": "Feature",
            "properties": self.properties,
            "geometry": {
                "type": self.RGJType,
                "coordinates": self.coordinates.tolist() if isinstance(self.coordinates, np.ndarray) else self.coordinates,
                "repulsion": self.repulsion.tolist()
            }
        }
    
    def __repr__(self):
        """
        Returns a more concise and unambiguous string representation of the object,
        typically used in debugging.
        """
        return str(self.toRGeoJSON())

    def __str__(self):
        """
        Returns a user-friendly string representation of the object,
        using its RGeoJSON representation.
        """

        def ndnumpy_to_str(array):
            string = f"{array.tolist()}"

            if len(string) > 50:
                string = string[:25] + "..." + string[-25:]

            return string

        return f"{self.__class__.__name__}(coordinates={ndnumpy_to_str(self.coordinates)} repulsion={ndnumpy_to_str(self.repulsion)})"


class PointRGJ(RGJGeometry):
    RGJType = "Point"

    def __init__(self, coordinates: Union[np.ndarray, Point], repulsion:Optional[np.ndarray] = None, **kwargs) -> None:
        super().__init__(coordinates=coordinates, repulsion=repulsion, **kwargs)

    def repulsion_vector(self, x: np.ndarray, **kwargs) -> np.ndarray:
        return x - self.coordinates

class LineStringRGJ(RGJGeometry):
    RGJType = "LineString"

    def __init__(self, coordinates: np.ndarray, repulsion:Optional[np.ndarray] = None, **kwargs) -> None:
        super().__init__(coordinates=coordinates, repulsion=repulsion, **kwargs)
        self.lines_n = len(self.coordinates) - 1
        self.points_in_line_pair = np.stack([self.coordinates[:-1], self.coordinates[1:]], axis=1)

    def set_coordinates(self, new_coords):
        super().set_coordinates(new_coords)
        self.lines_n = len(self.coordinates) - 1
        self.points_in_line_pair = np.stack([self.coordinates[:-1], self.coordinates[1:]], axis=1)
    
    def __repulsion_vector_one_line__(self, args) -> np.ndarray:
        x, line = args
        x2_d_x1 = line[1:2] - line[0:1]
        x_d_x1 = x - line[0:1]

        x12dotxx1 = (x2_d_x1*x_d_x1).sum(1, keepdims=True)
        x12dotx12 = (x2_d_x1*x2_d_x1).sum(1, keepdims=True)

        g = line[0] + np.clip(x12dotxx1/x12dotx12, 0.0, 1.0)*(x2_d_x1)
        return x - g
    
    def repulsion_vector(self, x: np.ndarray, min_dist_select:bool = True, **kwargs) -> np.ndarray:
        vectors:np.ndarray = [self.__repulsion_vector_one_line__((x, line)) for line in self.points_in_line_pair]

        vectors = np.stack(vectors, axis=0)
        if min_dist_select:
            vectors = vectors.swapaxes(0, 1)
            matrix = self.get_dist_matrix(scaled=True, inverted=True)
            nvectors = np.matmul(vectors, matrix)
            dist = (vectors*nvectors).sum(-1)
            select = dist.argmin(1)
            vectors = vectors[np.arange(len(select)), select]
        
        return vectors
    
class RectangleRGJ(RGJGeometry):
    RGJType = "Rectangle"

    def __init__(self, coordinates: np.ndarray, repulsion:Optional[np.ndarray] = None, **kwargs) -> None:
        super().__init__(coordinates=coordinates, repulsion=repulsion, **kwargs)
        self.x1_abs_x2 = np.abs(self.coordinates[0] - self.coordinates[1])

    def set_coordinates(self, new_coords):
        super().set_coordinates(new_coords)
        self.x1_abs_x2 = np.abs(self.coordinates[0] - self.coordinates[1])

    def repulsion_vector(self, x: np.ndarray, **kwargs) -> np.ndarray:
        
        return 0.5*np.sign(x-self.coordinates[0])*(np.abs(x-self.coordinates[0]) + np.abs(x-self.coordinates[1]) - self.x1_abs_x2)
    
class EllipseRGJ(RGJGeometry):
    RGJType = "Ellipse"
    DEN_ERROR_BUFFER = 1e-6

    def __init__(self, coordinates: np.ndarray, shape: np.ndarray, repulsion:Optional[np.ndarray] = None, **kwargs) -> None:

        super().__init__(coordinates=coordinates, repulsion=repulsion, **kwargs)
        self.shape = shape
        self.inv_shape = np.linalg.inv(self.shape)

        eval, evec  = np.linalg.eig(self.shape)
        vectors = evec * np.sqrt(eval) @ np.linalg.inv(evec)
        bounds = np.concatenate([self.coordinates + vectors, self.coordinates - vectors])
        self.bbox = np.array([bounds.min(0), bounds.max(0)])

    def set_coordinates(self, new_coords):
        super().set_coordinates(new_coords)

        eval, evec  = np.linalg.eig(self.shape)
        vectors = evec * np.sqrt(eval) @ np.linalg.inv(evec)
        bounds = np.concatenate([self.coordinates + vectors, self.coordinates - vectors])
        self.bbox = np.array([bounds.min(0), bounds.max(0)])

    def set_shape(self, new_shape):
        self.shape = new_shape
        self.inv_shape = np.linalg.inv(self.shape)

    def repulsion_vector(self, x: np.ndarray, **kwargs) -> np.ndarray:

        x_d_xh = x - self.coordinates
        Binvx = x_d_xh@self.inv_shape.T

        den = np.sqrt((Binvx*Binvx).sum(1, keepdims=True))
        den = np.maximum(den, self.DEN_ERROR_BUFFER)

        return np.maximum(1 - 1/den, 0)*x_d_xh
    
    def toRGeoJSON(self) -> RGeoJSONObject:
        out = super().toRGeoJSON()
        out["geometry"]['shape'] = self.shape.tolist() # type: ignore
        return out

class MultiPointRGJ(RGJGeometry):
    RGJType = "MultiPoint"

    def __init__(self, coordinates: np.ndarray, repulsion:Optional[np.ndarray] = None, **kwargs) -> None:
        super().__init__(coordinates=coordinates, repulsion=repulsion)
        self.bbox = self.coordinates.copy()

    def in_bbox(self, x:Point) -> bool:
        return any(self.bbox == x)

    def repulsion_vector(self, x: np.ndarray, min_dist_select:bool = True, **kwargs) -> np.ndarray:
        x = np.array(x)

        n = x.shape[0]
        m = self.coordinates.shape[0]

        x = np.tile(x, (m, 1, 1)).transpose(1, 0, 2)
        y = np.tile(self.coordinates, (n, 1, 1))
        diff = y - x

        if min_dist_select:
            matrix = self.get_dist_matrix(scaled=True, inverted=True)
            Adiff = np.matmul(diff, matrix)
            dist = (Adiff*diff).sum(-1)
            select = dist.argmin(1)
            diff = diff[np.arange(len(select)), select]
        else:
            diff = diff.swapaxes(0, 1)

        return diff
    
class MultiLineStringRGJ(LineStringRGJ):
    RGJType = "MultiLineString"

    def __init__(self, coordinates: np.ndarray, repulsion:Optional[np.ndarray] = None, properties:Optional[dict] = None, optional_dim = 2, **kwargs) -> None:

        self.coordinates = [np.array(coords) for coords in coordinates]
        self.repulsion = np.eye(optional_dim) if repulsion is None else np.array(repulsion)
        self.inv_repulsion = np.linalg.inv(self.repulsion)
        self.eye_repulsion = np.eye(len(self.repulsion))
        self.properties = {} if properties is None else properties
        self.grad_matrix = self.inv_repulsion + self.inv_repulsion.T
        
        self.lines_n = sum([len(coords)-1 for coords in self.coordinates])
        self.points_in_line_pair = np.concatenate([[coords[:-1], coords[1:]] for coords in self.coordinates], axis=1).swapaxes(0, 1)

        self.bbox = np.concatenate([np.array([coords.min(0), coords.max(0)]) for coords in self.coordinates])

    def set_coordinates(self, new_coords):
        self.coordinates = [np.array(coords) for coords in new_coords]
        self.lines_n = sum([len(coords)-1 for coords in self.coordinates])
        self.points_in_line_pair = np.concatenate([[coords[:-1], coords[1:]] for coords in self.coordinates], axis=1).swapaxes(0, 1)
        self.bbox = np.concatenate([np.array([coords.min(0), coords.max(0)]) for coords in self.coordinates])

    def get_center_point(self) -> np.ndarray:

        coords = np.concatenate([coords.reshape((-1, 2)) for coords in self.coordinates], axis=0)
        return (coords.min(0) + coords.max(0))/2.0

class MultiRectangleRGJ(RGJGeometry):

    RGJType = "MultiRectangle"

    def __init__(self, coordinates: np.ndarray, repulsion:Optional[np.ndarray] = None, **kwargs) -> None:
        super().__init__(coordinates=coordinates, repulsion=repulsion, **kwargs)
        self.rect_n = len(self.coordinates)
        self.bbox = np.array([np.array([coords.min(0), coords.max(0)]) for coords in self.coordinates])
        
    def set_coordinates(self, new_coords):
        super().set_coordinates(new_coords)
        self.rect_n = len(self.coordinates)
    
    def __repulsion_vector_one_rect__(self, args) -> np.ndarray:
        x, rect = args
        return 0.5*np.sign(x-rect[0])*(np.abs(x-rect[0]) + np.abs(x-rect[1]) - np.abs(rect[0] - rect[1]))
    
    def repulsion_vector(self, x: np.ndarray, min_dist_select:bool = True, **kwargs) -> np.ndarray:
        vectors:np.ndarray = [self.__repulsion_vector_one_rect__((x, rect)) for rect in self.coordinates]

        vectors = np.stack(vectors, axis=0)
        if min_dist_select:
            vectors = vectors.swapaxes(0, 1)
            matrix = self.get_dist_matrix(scaled=True, inverted=True)
            nvectors = np.matmul(vectors, matrix)
            dist = (vectors*nvectors).sum(-1)
            select = dist.argmin(1)
            vectors = vectors[np.arange(len(select)), select]
        
        return vectors

class MultiEllipseRGJ(RGJGeometry):
    RGJType = "MultiEllipse"
    DEN_ERROR_BUFFER = 1e-6

    def __init__(self, coordinates: np.ndarray, shape: np.ndarray, repulsion:Optional[np.ndarray] = None, **kwargs) -> None:
        super().__init__(coordinates=coordinates, repulsion=repulsion, **kwargs)
        self.shape = shape
        self.inv_shape = np.linalg.inv(self.shape)
        self.parameters = list(zip(self.coordinates, self.inv_shape))
        self.ellipse_n = len(self.coordinates)

        self.bbox = []

        for sp in self.shape:
            eval, evec  = np.linalg.eig(sp)
            vectors = evec * np.sqrt(eval) @ np.linalg.inv(evec)
            bounds = np.concatenate([self.coordinates + vectors, self.coordinates - vectors])
            self.bbox.append(np.array([bounds.min(0), bounds.max(0)]))

        self.bbox = np.array(self.bbox)

    def set_coordinates(self, new_coords):
        super().set_coordinates(new_coords)
        self.parameters = list(zip(self.coordinates, self.inv_shape))
        self.ellipse_n = len(self.coordinates)

        self.bbox = []
        for sp in self.shape:
            eval, evec  = np.linalg.eig(sp)
            vectors = evec * np.sqrt(eval) @ np.linalg.inv(evec)
            bounds = np.concatenate([self.coordinates + vectors, self.coordinates - vectors])
            self.bbox.append(np.array([bounds.min(0), bounds.max(0)]))

        self.bbox = np.array(self.bbox)

    def set_shape(self, new_shape):
        self.shape = new_shape
        self.inv_shape = np.linalg.inv(self.shape)
        self.parameters = list(zip(self.coordinates, self.inv_shape))
    
    def __repulsion_vector_one_ellipse__(self, args) -> np.ndarray:
        x, coordinate, inv_shape = args
        x_d_xh = x - coordinate
        Binvx = x_d_xh@inv_shape.T

        den = np.sqrt((Binvx*Binvx).sum(1, keepdims=True))
        den = np.maximum(den, self.DEN_ERROR_BUFFER)
        return np.maximum(1 - 1/den, 0)*x_d_xh
    
    def repulsion_vector(self, x: np.ndarray, min_dist_select:bool = True, **kwargs) -> np.ndarray:
        vectors:np.ndarray = [self.__repulsion_vector_one_ellipse__((x, parameters[0], parameters[1])) for parameters in self.parameters]

        vectors = np.stack(vectors, axis=0)
        if min_dist_select:
            vectors = vectors.swapaxes(0, 1)
            matrix = self.get_dist_matrix(scaled=True, inverted=True)
            nvectors = np.matmul(vectors, matrix) # Check: np.einsum('jk,ik->ij', matrix, vectors)
            dist = (vectors*nvectors).sum(-1)
            select = dist.argmin(1)
            vectors = vectors[np.arange(len(select)), select]
        
        return vectors
    
class GeometryCollectionRGJ(RGJGeometry):
    RGJType = "GeometryCollection"

    def __init__(self, geometries: List[RGJDict], properties:Optional[dict] = None, **kwargs) -> None:
        
        self.properties = {} if properties is None else properties
        self.rgjs:List[RGJGeometry] = [globals()[rgj_dict["type"]+"RGJ"](**rgj_dict) for rgj_dict in geometries]
        self.rgjs_n = len(self.rgjs)

        self.inv_repulsions = self.get_dist_matrix(scaled=True, inverted=True)
        self.grad_matrixes = np.array([rgj.grad_matrix for rgj in self.rgjs])

        bbox = np.concatenate([rgj.bbox for rgj in self.rgjs], 0).reshape(-1, 2)
        self.bbox = np.array([bbox.min(0), bbox.max(0)])

    def set_coordinates(self, new_coords):
        raise NotImplementedError

    def set_repulsion(self, new_repulsion):
        for rgj in self.rgjs:
            rgj.set_repulsion(new_repulsion)

    def in_bbox(self, x:Point) -> bool:
        return any([rgj.in_bbox(x) for rgj in self.rgjs])

    def get_dist_matrix(self, scaled=True, inverted=True) -> List[np.ndarray]:

        """
        Returns all distance matrix for all sub units
        """

        return np.array([rgj.get_dist_matrix(scaled=scaled, inverted=inverted) for rgj in self.rgjs])

    def get_center_point(self) -> np.ndarray:
        coords = np.reshape(np.array([rgj.get_center_point() for rgj in self.rgjs]), (-1, 2))
        return (coords.min(0) + coords.max(0))/2.0
    
    def squared_dist(self, x: np.ndarray, scaled=True, inverted=True, return_reference=False, **kwargs) -> np.ndarray:

        dists = np.stack([rgj.squared_dist(x, scaled=scaled, inverted=inverted) for rgj in self.rgjs], axis=1)

        if return_reference:
            min_idxs = np.argmin(dists, axis=1)
            return dists[np.arange(len(dists)), min_idxs], min_idxs

        return np.min(dists, axis=1)
    
    def repulsion_vector(self, x:np.ndarray, min_dist_select:bool = True, **kwargs) -> np.ndarray:

        vectors:np.ndarray = [rgj.repulsion_vector(x, min_dist_select=min_dist_select, **kwargs).reshape(-1, 2) for rgj in self.rgjs]

        vectors = np.stack(vectors, axis=0)

        if min_dist_select:
            vectors = vectors.swapaxes(0, 1)
            nvectors = np.einsum('ijk,lik->lij', self.inv_repulsions, vectors)
            dist = (vectors*nvectors).sum(-1)
            select = dist.argmin(1)
            vectors = vectors[np.arange(len(select)), select]

        return vectors
    
    def gradient(self, x: np.ndarray, **kwargs):

        _, dist_idxs = self.squared_dist(x, return_reference=True, **kwargs)

        repulsion_vector = self.repulsion_vector(x, min_dist_select = True, **kwargs)
        return - self.eval(x=x).reshape(-1, 1) * (np.einsum('ijk,ik->ij', self.grad_matrixes[dist_idxs], repulsion_vector))
    
    def toRGeoJSON(self) -> RGeoJSONObject:
        if self.RGJType is None: 
            return UserWarning(f"Object doesn't have a RGJType")
        
        return {
            "type": "Feature",
            "properties": self.properties,
            "geometry": {
                "type": self.RGJType,
                "geometries": [rgj.toRGeoJSON()["geometry"] for rgj in self.rgjs]
            }
        }
    
class PotentialField():
    
    """
    Represents a potential field composed of a subset of RGJ objects.

    This class allows the storage and spatial organization of RGJ geometries, and provides tools
    for bounding box management and automatic sizing/centering of the potential field based on input RGJs.

    Attributes:
        rgjs (List[RGJGeometry]): List of RGJ geometries in the field.
        center_point (Point or None): Central reference point of the potential field. Can be auto-calculated.
        size (np.ndarray or None): Spatial extent of the field (width, height). Can be inferred.
        extra_info (dict): Optional metadata or user-defined information.
        bbox (np.ndarray): Bounding box of all RGJs in the field. Shape (2, 2) -> [[xmin, ymin], [xmax, ymax]].

    Args:
        rgjs (Optional[List[RGJDict] or RGJGeometry]): List of RGJ objects or single RGJ geometry. Can be None.
        center_point (Optional[Point]): Optional center point for the field. If not provided, auto-calculated.
        size (Optional[FieldSize or float]): Width/height or 2D size of the field. If None, inferred from RGJs.
        properties (Optional[List[dict]]): List of metadata dictionaries for each RGJ (used when RGJs are dicts).
        extra_info (dict): Arbitrary extra user-defined metadata associated with the field.

    """

    def __init__(self, rgjs:Optional[Union[List[RGJDict], RGJGeometry]] = None, center_point: Optional[Point] = None, size:Optional[Union[FieldSize, float]] = None, properties:Optional[List[dict]] = None, extra_info={}):
        self.rgjs:List[RGJGeometry] = []
        self.__reload_center = None
        self.center_point = center_point
        self.extra_info = extra_info
        self.bbox = np.array([[None, None], [None, None]])

        if size is None:
            self.size = size
        elif np.isscalar(size):
            self.size = np.array([size, size])
        else:
            self.size = np.atleast_1d(size)

        if properties is None or isinstance(rgjs[0], RGJGeometry):
            for rgj in rgjs:
                self.addRGJ(rgj=rgj, reload_bbox=False)
        else:
            for rgj, proper in zip(rgjs, properties):
                self.addRGJ(rgj=rgj, properties=proper, reload_bbox=False)
        
        self.reload_bbox()

        if self.center_point is None:
            self.__reload_center = True # whether to recalculate center point if new RGJ are added
            if len(rgjs) > 0:
                self.center_point, suggest_size = self.__calculate_center_point__(suggest_size=True)
                self.size = np.array([max(suggest_size)]*2, dtype=float) if self.size is None else self.size
        else:
            self.__reload_center = False
            if len(rgjs) > 0:
                suggest_size = np.array([max(np.abs(self.bbox - self.center_point).reshape(-1))*2]*2)

                self.size = suggest_size if self.size is None else self.size

    def __getitem__(self, idxs: Union[int, np.integer, Iterable[int]]):
        """
        Access RGJ(s) in the field by index or list of indices.

        Args:
            idxs (int, np.integer, or Iterable[int]): Index or iterable of indices into the RGJ list.

        Returns:
            RGJGeometry or List[RGJGeometry] or None: The selected RGJ(s), or None if invalid input.
        """

        if isinstance(idxs, (int, np.integer)):
            return self.rgjs[int(idxs)]

        elif isinstance(idxs, Iterable) and not isinstance(idxs, (str, bytes)):
            return [self.rgjs[int(i)] for i in idxs]

        warnings.warn(f"Object of type {type(idxs)} not supported")
        return None

    def __iter__(self):
        self.rgj_idx = 0
        return self
    
    def __next__(self):
        if self.rgj_idx >= len(self):
            raise StopIteration
        out = self.rgjs[self.rgj_idx]
        self.rgj_idx += 1
        return out

    def __len__(self)->int:
        return len(self.rgjs)

    def __calculate_center_point__(self, suggest_size = False) -> Union[Point, Tuple[Point, float]]:
        center = np.sum(self.bbox, 0)/2.0

        if suggest_size:
            suggest_size = self.bbox[1] - self.bbox[0]
            return center, suggest_size
        
        return center
    
    def set_bbox(self, x_min:float, y_min:float, x_max:float, y_max:float):
        
        self.bbox = np.array([[x_min, y_min], [x_max, y_max]])
        self.center_point = np.sum(self.bbox, 0)/2.0
        self.size = self.bbox[1] - self.bbox[0]

    def set_all_repulsion(self, new_repulsion):
        new_repulsion = np.array(new_repulsion)
        for rgj in self.rgjs:
            rgj.set_repulsion(new_repulsion)

    def reload_bbox(self):
        if len(self):
            bbox = np.concatenate([rgj.bbox.reshape(-1, 2) for rgj in self.rgjs], 0)
            self.bbox = np.array([bbox.min(0), bbox.max(0)])
        else:
            self.bbox = np.array([[None, None], [None, None]])

        return self.bbox
    
    def reload_center_point(self, toggle=True, recal_size=False) -> Point:
        self.__reload_center = toggle
        if toggle and len(self.rgjs) > 0:
            if recal_size:
                self.center_point, suggest_size = self.__calculate_center_point__(True)
                self.size = np.array([max(suggest_size)]*2, dtype=float)
            else:
                self.center_point = self.__calculate_center_point__(False)

        return self.center_point

    def addRGJ(self, rgj:Union[RGJDict, RGJGeometry], properties:Optional[dict] = None, reload_bbox = True, **kward) -> None:

        if not isinstance(rgj, RGJGeometry):
            if not isinstance(rgj, dict) or "type" not in rgj:
                raise ValueError("RGJ must be an RGJGeometry.")
            
            cls = globals().get(rgj["type"] + "RGJ")
            if cls is None:
                raise ValueError(f"No RGJ class found for type {rgj['type']}")
            
            rgj = cls(properties=properties, **rgj, **kward)
        
        self.rgjs.append(rgj)

        if reload_bbox:
            self.reload_bbox()

        if self.__reload_center:
            self.center_point = self.__calculate_center_point__()

    def addField(self, new_field:PotentialField, reload_bbox = True):

        """
        Adds all RGJs from another potential field to this field.

        Args:
            new_field (PotentialField): A potential field containing RGJs to add.
            reload_bbox (bool, optional): Whether to recompute the bounding box after addition.
                                        Defaults to True.
        """
        og_reload_center = self.__reload_center
        self.__reload_center = False
        
        # Add rgj to field
        try:
            for rgj in new_field:
                self.addRGJ(rgj=rgj, reload_bbox=False)
        finally:
            self.__reload_center = og_reload_center

        if reload_bbox:
            self.reload_bbox()

        if self.__reload_center:
            self.center_point = self.__calculate_center_point__()

    def delRGJ(self, idxs: Union[int, List[int], np.ndarray], reload_bbox: bool = True) -> None:
        """
        Removes one or more RGJs from the field by their indices.

        Args:
            idxs (Union[int, List[int], np.ndarray]): A single index, list of indices, or 1D numpy array of indices to remove.
            reload_bbox (bool, optional): Whether to recompute the bounding box after deletion. Defaults to True.
        """
        # Normalize input and wrap indices within range
        idxs = np.atleast_1d(idxs).astype(int)
        idxs = np.unique(idxs % len(self))[::-1]  # Wrap, deduplicate, and reverse sort

        # Delete RGJs from highest to lowest index to avoid reindexing errors
        for idx in idxs:
            del self.rgjs[idx]

        # Optionally recalculate the center point if enabled
        if self.__reload_center:
            self.center_point = self.__calculate_center_point__()

        # Optionally recompute the bounding box
        if reload_bbox:
            self.reload_bbox()
   
    def in_bbox(self, point: Point, filted_idx: Optional[List[int]] = None) -> bool:
        """
        Check if the point lies within the bounding box of any RGJ.

        Args:
            point (Point): The 2D point to check.
            filted_idx (Optional[List[int]]): Optional subset of RGJ indices to filter.

        Returns:
            bool: True if the point is in any RGJ's bounding box.
        """
        point = np.array(point)
        rgjs = [self.rgjs[idx] for idx in filted_idx] if filted_idx is not None else self.rgjs

        return any(rgj.in_bbox(point) for rgj in rgjs)
    
    def find_bbox(self, point: Point, filted_idx: Optional[List[int]] = None) -> np.ndarray:
        """
        Return indices of RGJs whose bounding boxes contain the given point.

        Args:
            point (Point): The 2D point to check.
            filted_idx (Optional[List[int]]): Optional subset of RGJ indices to filter.

        Returns:
            np.ndarray: Indices of RGJs (in global index space) containing the point.
        """
        point = np.array(point)

        if filted_idx is not None:
            return np.array([idx for idx in filted_idx if self.rgjs[idx].in_bbox(point)])
        else:
            return np.nonzero([rgj.in_bbox(point) for rgj in self.rgjs])[0]
    
    def repulsion_vectors(self, points: Union[np.ndarray, List[Point]], filted_idx:Optional[List[int]] = None, min_dist_select:bool = True, return_reference = False) -> Union[np.ndarray, RepulsionVectorsAndRef]:
        points = np.atleast_2d(points).astype(float)
        if not len(self):
            return points*np.inf
        filted_idx = filted_idx if not filted_idx is None else list(range(len(self)))

        if return_reference:
            idxs = []
            repulsion_vectors = []

            for idx in filted_idx:
                vectors = self.rgjs[idx].repulsion_vector(points, min_dist_select=min_dist_select).reshape(-1, 2)

                idxs.extend([idx]*len(vectors))
                repulsion_vectors.append(vectors)

            repulsion_vectors = np.concatenate(repulsion_vectors, axis=0)
            return repulsion_vectors, np.array(idxs, dtype=int)
   
        else:
            rgjs = [self.rgjs[idx] for idx in filted_idx]
            return np.concatenate([rgj.repulsion_vector(points, min_dist_select=min_dist_select).reshape(-1, 2) for rgj in rgjs], axis=0)
    
    def gradient(self, points: Union[np.ndarray, List[Point]], min_dist_select=True) -> np.ndarray:

        points = np.atleast_2d(points).astype(float)
        if not len(self):
            return points * 0.0
        
        grad = np.zeros((len(points), 2), dtype=float)

        # Use closest RGJ per point to compute gradient
        _, grad_idxs = self.squared_dist(points=points, return_reference=True)
        unique_idxs = set(grad_idxs)
        
        for idx in unique_idxs:
            select = (grad_idxs == idx)
            grad[select] = self.rgjs[idx].gradient(points[select], min_dist_select=min_dist_select)
        
        return grad

    def eval(self, points: Union[np.ndarray, List[Point]], filted_idx:Optional[List[int]] = None) -> np.ndarray:
        points = np.atleast_2d(points).astype(float)
        rgjs = [self.rgjs[idx] for idx in filted_idx] if not filted_idx is None else self.rgjs

        if not len(rgjs):
            return points.sum(1)*0.0
        
        return np.max(np.stack([rgj.eval(points) for rgj in rgjs], axis=1), axis=1)
    
    def eval_per(self, points: Union[np.ndarray, List[Point]], idxs:Optional[List[int]] = None) -> np.ndarray:
        if len(points) != len(idxs):
            raise RuntimeError("The number of points doesn't match the number of indexes")
        
        points = np.atleast_2d(points).astype(float)
        n = len(points)
        idxs = np.array(idxs, dtype=int)
        
        evals = np.ones(n, dtype=points[0].dtype)
        for idx in set(idxs):
            select = idx == idxs
            evals[select] = self.rgjs[idx].eval(points[select])

        return evals
    
    def squared_dist(self, points:Union[np.ndarray, List[Point]], filted_idx:Optional[List[int]] = None, scaled=True, inverted=True, return_reference = False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        points = np.atleast_2d(points).astype(float)
        if not len(self):
            warnings.warn("There are not any RGJs elements in the field")
            if return_reference:
                return points.sum(1)*np.inf, -np.ones_like(points.sum(1))
            return points.sum(1)*np.inf

        dists = self.squared_dist_list(points=points, filted_idx=filted_idx, scaled=scaled, inverted=inverted)

        if return_reference:
            min_idxs = np.argmin(dists, axis=1)

            filted_idx = filted_idx if filted_idx is not None else np.arange(len(self))
            return dists[np.arange(len(dists)), min_idxs], filted_idx[min_idxs]

        return np.min(dists, axis=1)
    
    def squared_dist_per(self, points: Union[np.ndarray, List[Point]], idxs:Optional[List[int]] = None, scaled=True, inverted=True) -> np.ndarray:

        idxs = [] if idxs is None else idxs
        n = len(points)

        if n != len(idxs):
            if not len(idxs):
                raise RuntimeError("The number of points doesn't match the number of indexes")
            else:
                warnings.warn("The number of points doesn't match the number of indexes. Each point matched with each rgj")
                idxs = np.arange(n)
        
        points = np.array(points)
        idxs = np.array(idxs, dtype=int)
        
        dists = np.ones(n, dtype=points[0].dtype)
        for idx in set(idxs):
            select = idx == idxs
            dists[select] = self.rgjs[idx].squared_dist(points[select], scaled=scaled, inverted=inverted)

        return dists
    
    def squared_dist_list(self, points:Union[np.ndarray, List[Point]], filted_idx:Optional[List[int]] = None, scaled=True, inverted=True) -> np.ndarray:
        points = np.atleast_2d(points).astype(float)
        rgjs = [self.rgjs[idx] for idx in filted_idx] if not filted_idx is None else self.rgjs

        if not len(rgjs):
            warnings.warn("There are not any RGJs elements in the field")
            return np.ones((len(points), len(rgjs)))*np.inf

        return np.stack([rgj.squared_dist(points, scaled=scaled, inverted=inverted) for rgj in rgjs], axis=1)
    
    def estimate_route_area(self, route:Union[List[Point], np.ndarray], step=1e-3, n=0, scale_transform:Scaler = lambda x: x) -> float:
        route = np.array(route)

        points, step, _ = lpf.interpolate_along_route(route=route, step=step, n=n, return_step_n=True)
        points = points if n <= 0 else points[:-1]

        f_eval = scale_transform(self.eval(points=points))

        return f_eval.sum()*step
    
    def estimate_route_highest_potential(self, route:Union[List[Point], np.ndarray], step=1e-2, n=0, scale_transform:Scaler = lambda x: x) -> float:
        route = np.array(route)

        points, step, _ = lpf.interpolate_along_route(route=route, step=step, n=n, return_step_n=True)
        points = points if n <= 0 else points[:-1]

        f_eval:np.ndarray = scale_transform(self.eval(points=points))

        return f_eval.max()

    def to_image(self, resolution:int = 200, margin:float = 0.0, center_point:Optional[Point] = None, size:Optional[FieldSize] = None, filted_idx:Optional[List[int]] = None, return_extent=True) -> Union[np.ndarray, Tuple[np.ndarray, List[float]]]:

        if center_point is None:
            if self.center_point is None:
                raise RuntimeError('Center point for field has not been defined')
            
            center_point = self.center_point

        if size is None:
            if self.size is None:
                raise RuntimeError('Size of field has not been defined')

            size = self.size
        else:
            size = np.array(size)

        n2 = size/2.0

        loc_tl = np.array(center_point) + np.array([-n2[0]-margin, n2[1]+margin])
        loc_br = np.array(center_point) + np.array([n2[0]+margin, -n2[1]-margin])

        y_resolution = int(resolution*abs(loc_tl[1] - loc_br[1])/abs(loc_br[0] - loc_tl[0]))
        xaxis = np.linspace(loc_tl[0], loc_br[0], resolution)
        yaxis = np.linspace(loc_tl[1], loc_br[1], y_resolution)

        xgrid, ygrid = np.meshgrid(xaxis, yaxis)
        points = np.vstack([xgrid.ravel(), ygrid.ravel()]).T

        image = self.eval(points, filted_idx=filted_idx).reshape((y_resolution, resolution))

        if return_extent:
            extent = np.reshape([loc_tl[0], loc_br[0], loc_br[1], loc_tl[1]], -1).tolist()
            return image, extent

        return image
    
    def toRGeoJSON(self, return_bbox=False) -> RGeoJSONCollection:\
    
        rgeojson = {
            'type': 'FeatureCollection',
            '_version_': "2D",
            'features': [rgj.toRGeoJSON() for rgj in self.rgjs],
            **self.extra_info
        }
        if return_bbox:
            extent = self.get_extent()
            rgeojson["bbox"] = extent[::2] + extent[1::2]

        return rgeojson
 